
import numpy as np
import math

import netCDF4
import datetime

import hashlib

import copy

import os

import cdb_query_archive_class

import json

import retrieval_utils

import indices_utils

import multiprocessing
import subprocess

import netcdf_soft_links

import nc_Database

def get_year_axis(path_name):
    try:
        #print 'Loading file... ',
        #print path_name
        data=netCDF4.Dataset(path_name)
        dimensions_list=data.dimensions.keys()
        if 'time' not in dimensions_list:
            raise Error('time is missing from variable')
        date_axis = get_date_axis(data.variables['time'])
        #print ' Done!'
        data.close()
        year_axis=np.array([date.year for date in date_axis])
        month_axis=np.array([date.month for date in date_axis])
    except:
        return None, None

    return year_axis, month_axis

def get_date_axis(time_var):
    units=time_var.units
    if 'calendar' in dir(time_var):
        calendar=time_var.calendar
    else:
        calendar=None
    return get_date_axis_units(time_var[:],units,calendar)

def get_date_axis_units(time_axis,units,calendar):
    if units=='day as %Y%m%d.%f':
        date_axis=get_date_axis_absolute(time_axis)
    else:
        date_axis=get_date_axis_relative(time_axis,units,calendar)
    return date_axis

def get_date_axis_relative(time_axis,units,calendar):
    if calendar!=None:
        date_axis = netCDF4.num2date(time_axis,units=units,calendar=calendar)
    else:
        date_axis = netCDF4.num2date(time_axis,units=units)
    return date_axis

def get_date_axis_absolute(time_axis):
    return map(convert_to_date_absolute,time_axis)

def convert_to_date_absolute(absolute_time):
    year=int(math.floor(absolute_time/1e4))
    remainder=absolute_time-year*1e4
    month=int(math.floor(remainder/1e2))
    remainder-=month*1e2
    day=int(math.floor(remainder))
    remainder-=day
    remainder*=24.0
    hour=int(math.floor(remainder))
    remainder-=hour
    remainder*=60.0
    minute=int(math.floor(remainder))
    remainder-=minute
    remainder*=60.0
    seconds=int(math.floor(remainder))
    return datetime.datetime(year,month,day,hour,minute,seconds)

#def replicate_full_netcdf_recursive(output,data,options=None):
def replicate_full_netcdf_recursive(output,data,check_empty=False):
    for var_name in data.variables.keys():
        replicate_and_copy_variable(output,data,var_name,check_empty=check_empty)
    if len(data.groups.keys())>0:
        for group in data.groups.keys():
            output_grp=replicate_group(output,data,group)
            replicate_full_netcdf_recursive(output_grp,data.groups[group],check_empty=check_empty)
    return

def replicate_and_copy_variable(output,data,var_name,datatype=None,fill_value=None,add_dim=None,chunksize=None,zlib=None,check_empty=False):
    replicate_netcdf_var(output,data,var_name,datatype=datatype,fill_value=fill_value,add_dim=add_dim,chunksize=chunksize,zlib=zlib)
    if len(data.variables[var_name].shape)>0 and min(data.variables[var_name].shape)>0:
        max_request=4500 #maximum request in Mb
        max_time_steps=max(
                        int(np.floor(max_request*1024*1024/(32*np.prod(data.variables[var_name].shape)))),
                        1)

        num_time_chunk=int(np.ceil(data.variables[var_name].shape[0]/float(max_time_steps)))

        if (len(data.variables[var_name].shape)>1 and
            num_time_chunk>1):
            for time_chunk in range(num_time_chunk):
                time_slice=slice(time_chunk*max_time_steps,
                                 min((time_chunk+1)*max_time_steps,data.variables[var_name].shape[0])
                                 ,1)
                                 #(time_chunk+1)*max_time_steps,1)
                #output.variables[var_name][time_slice,...]=data.variables[var_name][time_slice,...]
                temp=data.variables[var_name][time_slice,...]
                #Assign only if not masked everywhere:
                #if not ('mask' in dir(temp) and temp.mask.all()):
                #output.variables[var_name][time_slice,...]=temp
                if not 'mask' in dir(temp) or not check_empty:
                    output.variables[var_name][time_slice,...]=temp
                else: 
                    if not temp.mask.all():
                        output.variables[var_name][time_slice,...]=temp
                output.sync()
        else:
            #output.variables[var_name][:]=data.variables[var_name][:]
            temp=data.variables[var_name][:]
            if not 'mask' in dir(temp) or not check_empty:
                output.variables[var_name][:]=temp
            else: 
                if not temp.mask.all():
                    output.variables[var_name][:]=temp
    elif len(data.variables[var_name].dimensions)==0:
        #scalar variable:
        output.variables[var_name][:]=data.variables[var_name][:]
    return output

def replicate_group(output,data,group_name):
    output_grp=create_group(output,data,group_name)
    replicate_netcdf_file(output_grp,data.groups[group_name])
    return output_grp

def create_group(output,data,group_name):
    if not group_name in output.groups.keys():
        output_grp=output.createGroup(group_name)
    else:
        output_grp=output.groups[group_name]
    return output_grp
    
def replicate_netcdf_file(output,data):
    for att in data.ncattrs():
        att_val=data.getncattr(att)
        if 'encode' in dir(att_val):
            att_val=att_val.encode('ascii','replace')
        if (not att in output.ncattrs() and
            att != 'cdb_query_temp'):
                output.setncattr(att,att_val)
    return output

def replicate_netcdf_var_dimensions(output,data,var):
    for dims in data.variables[var].dimensions:
        if dims not in output.dimensions.keys() and dims in data.dimensions.keys():
            if data.dimensions[dims].isunlimited():
                output.createDimension(dims,None)
            else:
                output.createDimension(dims,len(data.dimensions[dims]))
            if dims in data.variables.keys():
                dim_var = output.createVariable(dims,data.variables[dims].dtype,(dims,))
                dim_var[:] = data.variables[dims][:]
                output = replicate_netcdf_var(output,data,dims)
                if 'bounds' in output.variables[dims].ncattrs():
                    output=replicate_netcdf_var(output,data,output.variables[dims].getncattr('bounds'))
                    output.variables[output.variables[dims].getncattr('bounds')][:]=data.variables[output.variables[dims].getncattr('bounds')][:]
            else:
                #Create a dummy dimension variable:
                dim_var = output.createVariable(dims,np.float,(dims,))
                dim_var[:]=np.arange(len(data.dimensions[dims]))
    return output

def replicate_netcdf_var(output,data,var,datatype=None,fill_value=None,add_dim=None,chunksize=None,zlib=None):
    output=replicate_netcdf_var_dimensions(output,data,var)
    if datatype==None: datatype=data.variables[var].datatype
    if (isinstance(datatype,netCDF4.CompoundType) and
        not datatype.name in output.cmptypes.keys()):
        datatype=output.createCompoundType(datatype.dtype,datatype.name)

    kwargs=dict()
    if (fill_value==None and 
        '_FillValue' in dir(data.variables[var]) and 
        datatype==data.variables[var].datatype):
            kwargs['fill_value']=data.variables[var]._FillValue
    else:
        kwargs['fill_value']=fill_value

    if zlib==None:
        if data.variables[var].filters()==None:
            kwargs['zlib']=False
        else:
            for item in data.variables[var].filters():
                kwargs[item]=data.variables[var].filters()[item]
    else:
        kwargs['zlib']=zlib
    
    if var not in output.variables.keys():
        dimensions=data.variables[var].dimensions
        if add_dim:
            dimensions+=(add_dim,)
        if chunksize!=None:
            if chunksize==-1:
                chunksizes=tuple([1 if dim=='time' else data.variables[var].shape[dim_id] for dim_id,dim in enumerate(dimensions)])
            else:
                #chunksizes=tuple([1 if output.dimensions[dim].isunlimited() else 10 for dim in dimensions])
                #chunksizes=tuple([1 if dim=='time' else chunksize for dim in dimensions])
                chunksizes=tuple([1 if dim=='time' else chunksize for dim_id,dim in enumerate(dimensions)])
            kwargs['chunksizes']=chunksizes
        else:
            if data.variables[var].chunking()=='contiguous':
                kwargs['contiguous']=True
            else:
                kwargs['chunksizes']=data.variables[var].chunking()
        output.createVariable(var,datatype,dimensions,**kwargs)
    output = replicate_netcdf_var_att(output,data,var)
    return output

def replicate_netcdf_var_att(output,data,var):
    for att in data.variables[var].ncattrs():
        att_val=data.variables[var].getncattr(att)
        if att[0]!='_':
            if 'encode' in dir(att_val):
                att_val=att_val.encode('ascii','replace')
            output.variables[var].setncattr(att,att_val)
    return output

def create_time_axis(output,data,time_axis):
    #output.createDimension('time',len(time_axis))
    output.createDimension('time',None)
    time = output.createVariable('time','d',('time',))
    time.calendar=str(data.variables['time'].calendar)
    time.units=str(data.variables['time'].units)
    time[:]=time_axis
    return

def netcdf_calendar(data):
    if 'calendar' in dir(data.variables['time']):
        calendar=data.variables['time'].calendar
    else:
        calendar='standard'
    return calendar

def convert(options,project_drs):
    database=cdb_query_archive_class.SimpleTree(options,project_drs)
    #Recover the database meta data:
    vars_list=database.list_fields_local(options,database.drs.official_drs_no_version)
    database.load_header(options)

    #Find the list of arguments to pass to convert_variable:
    args_list=[]
    for var_id,var in enumerate(vars_list):
        options_copy=copy.copy(options)
        for opt_id, opt in enumerate(database.drs.official_drs_no_version):
            setattr(options_copy,opt,var[opt_id])
        args_list.append((copy.copy(database),options_copy))
    if options.num_procs>1:
        pool=multiprocessing.Pool(processes=options.num_procs,maxtasksperchild=1)
        result=pool.map(convert_to_variable_tuple,args_list,chunksize=1)
        pool.close()
        pool.join()
    else:
        result=map(convert_to_variable_tuple,args_list)
    return

def convert_to_variable_tuple(x):
    return convert_to_variable(*x)

def convert_to_variable(database,options):
    input_file_name=options.in_diagnostic_netcdf_file

    options.version='v'+datetime.datetime.now().strftime('%Y%m%d')
    var=[getattr(options,opt) for opt in database.drs.official_drs]
    file_name=[getattr(options,opt) for opt in database.drs.filename_drs]
    output_file_name=options.out_destination+'/'+'/'.join(var)+'/'+'_'.join(file_name)
    temp_output_file_name=output_file_name+'.pid'+str(os.getpid())

    try:
        directory=os.path.dirname(temp_output_file_name)
        if not os.path.exists(directory):
            os.makedirs(directory)
    except:
        pass

    data=netCDF4.Dataset(input_file_name,'r')
    output_tmp=netCDF4.Dataset(temp_output_file_name,'w',format='NETCDF4',diskless=True,persist=True)
    #extract_netcdf_variable_recursive(output_tmp,data,options)
    #tree=zip(database.drs.official_drs_no_version,var)
    var=[getattr(options,opt) for opt in database.drs.official_drs_no_version]
    tree=zip(database.drs.official_drs_no_version,var)
    extract_netcdf_variable_recursive(output_tmp,data,tree[0],tree[1:],options)
    data.close()

    #Get the time:
    timestamp=convert_dates_to_timestamps(output_tmp,options.time_frequency)
    output_tmp.close()
    os.rename(temp_output_file_name,output_file_name+timestamp+'.nc')
    return

def convert_dates_to_timestamps(output_tmp,time_frequency):
    conversion=dict()
    conversion['year']=(lambda x: str(x.year).zfill(4))
    conversion['mon']=(lambda x: str(x.year).zfill(4)+str(x.month).zfill(2))
    conversion['day']=(lambda x: str(x.year).zfill(4)+str(x.month).zfill(2)+str(x.day).zfill(2))
    conversion['6hr']=(lambda x: str(x.year).zfill(4)+str(x.month).zfill(2)+str(x.day).zfill(2)+str(x.hour).zfill(2)+str(x.minute).zfill(2)+str(x.second).zfill(2))
    conversion['3hr']=(lambda x: str(x.year).zfill(4)+str(x.month).zfill(2)+str(x.day).zfill(2)+str(x.hour).zfill(2)+str(x.minute).zfill(2)+str(x.second).zfill(2))
    if time_frequency!='fx':
        date_axis=get_date_axis(output_tmp.variables['time'])[[0,-1]]
        return '_'+'-'.join([conversion[time_frequency](date) for date in date_axis])
    else:
        return ''


def assign_tree(output,val,sort_table,tree):
    if len(tree)>1:
        assign_tree(output.groups[tree[0]],val,sort_table,tree[1:])
    else:
        output.variables[tree[0]][sort_table]=val
    return

def apply(options,project_drs):
    if options.script=='': return
    database=cdb_query_archive_class.SimpleTree(options,project_drs)
    #Recover the database meta data:
    if options.field!=None:
        drs_to_eliminate=[field for field in database.drs.official_drs_no_version if
                                             not field in options.field]
    else:
        drs_to_eliminate=database.drs.official_drs_no_version
    vars_list=[[ var[drs_to_eliminate.index(field)] if field in drs_to_eliminate else None
                        for field in database.drs.official_drs_no_version] for var in 
                        database.list_fields_local(options,drs_to_eliminate) ]

    #Always add the fixed variables to apply:
    #vars_list=[ var for var in vars_list if var[database.drs.official_drs_no_version.index('ensemble')] !='r0i0p0' ]

    output_root=distributed_apply(apply_to_variable,database,options,vars_list)
    database.record_header(options,output_root)
    output_root.close()
    return

def apply_to_variable(database,options):
    input_file_name=options.in_diagnostic_netcdf_file
    files_list=[input_file_name,]+options.in_extra_netcdf_files

    output_file_name=options.out_netcdf_file+'.pid'+str(os.getpid())
    temp_output_file_name= output_file_name+'.tmp'

    var=[getattr(options,opt) for opt in database.drs.official_drs_no_version]
    tree=zip(database.drs.official_drs_no_version,var)

    if options.add_fixed:
        #Specification for fixed vars:
        var_fx=[ getattr(options,opt) if not opt in database.drs.var_specs+['var',] else None for opt in database.drs.official_drs_no_version]
        var_fx=copy.copy(var)
        var_fx[database.drs.official_drs_no_version.index('ensemble')]='r0i0p0'
        var_fx[database.drs.official_drs_no_version.index('var')]=None
        for opt  in database.drs.var_specs+['var',]:
            if opt in ['time_frequency','cmor_table']:
                var_fx[database.drs.official_drs_no_version.index(opt)]='fx'
        tree_fx=zip(database.drs.official_drs_no_version,var_fx)
        options_fx=copy.copy(options)
        for opt_id,opt in enumerate(tree_fx):
            if opt!=tree[opt_id]:
                setattr(options_fx,opt[0],opt[1])

    temp_files_list=[]
    for file in files_list:
        data=netCDF4.Dataset(file,'r')
        temp_file=file+'.pid'+str(os.getpid())
        output_tmp=netCDF4.Dataset(temp_file,'w',format='NETCDF4',diskless=True,persist=True)
        extract_netcdf_variable_recursive(output_tmp,data,tree[0],tree[1:],options,check_empty=True)
        if options.add_fixed:
            extract_netcdf_variable_recursive(output_tmp,data,tree_fx[0],tree_fx[1:],options_fx,check_empty=True)
        temp_files_list.append(temp_file)
        output_tmp.close()
        data.close()

    temp_files_list.append(temp_output_file_name)

    script_to_call=options.script
    for file_id, file in enumerate(temp_files_list):
        if not '{'+str(file_id)+'}' in options.script:
            script_to_call+=' {'+str(file_id)+'}'

    #script_to_call+=' '+temp_output_file_name

    #print '/'.join(var)
    #print script_to_call.format(*temp_files_list)
    out=subprocess.call(script_to_call.format(*temp_files_list),shell=True)

    #import shlex
    #args = shlex.split(script_to_call.format(*temp_files_list))
    #out=subprocess.call(args,shell=True)
    #out=subprocess.Popen(script_to_call.format(*temp_files_list),shell=True,
    #                     stdout=subprocess.PIPE,
    #                     stderr=subprocess.PIPE)
    #out.wait()

    try:
        for file in temp_files_list[:-1]:
            os.remove(file)
    except OSError:
        pass
    return (temp_output_file_name, var)

def extract_netcdf_variable_recursive(output,data,level_desc,tree,options,check_empty=False):
    level_name=level_desc[0]
    group_name=level_desc[1]
    if group_name==None or isinstance(group_name,list):
        for group in data.groups.keys():
            if ( nc_Database.is_level_name_included_and_not_excluded(level_name,options,group) and
                 nc_Database.retrieve_tree_recursive_check_not_empty(options,data.groups[group])):
                output_grp=replicate_group(output,data,group)
                if len(tree)>0:
                    extract_netcdf_variable_recursive(output_grp,data.groups[group],tree[0],tree[1:],options,check_empty=check_empty)
                else:
                    netcdf_pointers=netcdf_soft_links.read_netCDF_pointers(data.groups[group])
                    netcdf_pointers.replicate(output_grp,check_empty=check_empty)
    else:
        if len(tree)>0:
            extract_netcdf_variable_recursive(output,data.groups[group_name],tree[0],tree[1:],options,check_empty=check_empty)
        else:
            netcdf_pointers=netcdf_soft_links.read_netCDF_pointers(data.groups[group_name])
            netcdf_pointers.replicate(output,check_empty=check_empty)
    return


def worker(tuple):
    result=tuple[0](*tuple[1:-1])
    tuple[-1].put(result)
    return
    
def distributed_apply(function_handle,database,options,vars_list,args=tuple()):

        #Open output file:
        output_file_name=options.out_netcdf_file
        output_root=netCDF4.Dataset(output_file_name,'w',format='NETCDF4')

        manager=multiprocessing.Manager()
        queue=manager.Queue()
        #queue=multiprocessing.Queue()
        #Set up the discovery var per var:
        args_list=[]
        for var_id,var in enumerate(vars_list):
            options_copy=copy.copy(options)
            for opt_id, opt in enumerate(database.drs.official_drs_no_version):
                if var[opt_id]!=None:
                    setattr(options_copy,opt,var[opt_id])
            args_list.append((function_handle,copy.copy(database),options_copy)+args+(queue,))
        
        #Span up to options.num_procs processes and each child process analyzes only one simulation
        if options.num_procs>1:
            pool=multiprocessing.Pool(processes=options.num_procs,maxtasksperchild=1)
            result=pool.map_async(worker,args_list,chunksize=1)
            #Record files to main file:
            for arg in args_list:
                description=queue.get()
                temp_file_name=description[0]
                var=description[1]
                data=netCDF4.Dataset(temp_file_name,'r')
                tree=zip(database.drs.official_drs_no_version,var)
                replace_netcdf_variable_recursive(output_root,data,tree[0],tree[1:],check_empty=True)
                data.close()
                try:
                    os.remove(temp_file_name)
                except OSError:
                    pass
                output_root.sync()
    
            pool.close()
            pool.join()
        else:
            for arg in args_list:
                worker(arg)
                temp_file_name, var=queue.get()
                data=netCDF4.Dataset(temp_file_name,'r')
                tree=zip(database.drs.official_drs_no_version,var)
                replace_netcdf_variable_recursive(output_root,data,tree[0],tree[1:],check_empty=True)
                data.close()
                try:
                    os.remove(temp_file_name)
                except OSError:
                    pass
                output_root.sync()


        return output_root

def replace_netcdf_variable_recursive(output,data,level_desc,tree,check_empty=False):
    level_name=level_desc[0]
    group_name=level_desc[1]
    if group_name==None or isinstance(group_name,list):
        for group in data.groups.keys():
            output_grp=create_group(output,data,group)
            output_grp.setncattr('level_name',level_name)
            if len(tree)>0:
                    replace_netcdf_variable_recursive(output_grp,data.groups[group],tree[0],tree[1:],check_empty=check_empty)
            else:
                netcdf_pointers=netcdf_soft_links.read_netCDF_pointers(data.groups[group])
                netcdf_pointers.replicate(output_grp,check_empty=check_empty)
    else:
        output_grp=create_group(output,data,group_name)
        output_grp.setncattr('level_name',level_name)
        if len(tree)>0:
            replace_netcdf_variable_recursive(output_grp,data,tree[0],tree[1:],check_empty=check_empty)
        else:
            netcdf_pointers=netcdf_soft_links.read_netCDF_pointers(data)
            netcdf_pointers.replicate(output_grp,check_empty=check_empty)
        #    netcdf_pointers=netcdf_soft_links.read_netCDF_pointers(data)
        #    netcdf_pointers.replicate(output,check_empty=check_empty)
    return

    




import numpy as np

import netCDF4
import os

import json

import netcdf_utils

import remote_netcdf

import create_soft_links

import indices_utils

import retrieval_utils

import copy

import nc_Database

class create_netCDF_pointers:
    def __init__(self,file_type_list,data_node_list,semaphores=None):
        self.file_type_list=file_type_list
        self.data_node_list=data_node_list
        self.semaphores=semaphores
        return


    def record_paths(self,output,paths_list,var,years,months):
        #First order source files by preference:
        soft_links=create_soft_links.soft_links(paths_list,self.file_type_list,self.data_node_list,semaphores=self.semaphores)
        soft_links.create(output)
        return

    def record_meta_data(self,output,paths_list,var,years,months):
        #Retrieve time and meta:
        soft_links=create_soft_links.soft_links(paths_list,self.file_type_list,self.data_node_list,semaphores=self.semaphores)
        soft_links.create_variable(output,var,years,months)
        #Put version:
        output.setncattr('netcdf_soft_links_version','1.0')
        return

    def record_fx(self,output,paths_list,var):
        #Find the most recent version:
        most_recent_version='v'+str(np.max([int(item['version'][1:]) for item in paths_list]))
        path=paths_list[[item['version'] for item in paths_list].index(most_recent_version)]

        #Check if data in available:
        remote_data=remote_netcdf.remote_netCDF(path['path'].split('|')[0].replace('fileServer','dodsC'),self.semaphores)
        alt_path_name=remote_data.check_if_available_and_find_alternative([item['path'].split('|')[0].replace('fileServer','dodsC') for item in paths_list],
                                                                 [item['path'].split('|')[1] for item in paths_list]).replace('dodsC','fileServer')

        #Use aternative path:
        path=paths_list[[item['path'].split('|')[0] for item in paths_list].index(alt_path_name)]

        for att in path.keys():
            if att!='path':      
                output.setncattr(att,path[att])
        output.setncattr('path',path['path'].split('|')[0])
        output.setncattr('checksum',path['path'].split('|')[1])

        remote_data=remote_netcdf.remote_netCDF(path['path'].split('|')[0].replace('fileServer','dodsC'),self.semaphores)
        #open and record:
        try:
            remote_data.open_with_error()
            for var_name in remote_data.Dataset.variables.keys():
                netcdf_utils.replicate_and_copy_variable(output,remote_data.Dataset,var_name,zlib=True)
                #netcdf_utils.replicate_netcdf_var(output,remote_data.Dataset,var_name)
                #output.variables[var_name][:]=remote_data.Dataset.variables[var_name][:]
            output.sync()
        except dodsError as e:
            e_mod=" This is an uncommon error. It is likely to be FATAL."
            print e.value+e_mod
        remote_data.close()
        return


class read_netCDF_pointers:
    def __init__(self,data_root,Xdata_node=None,data_node=None,queues=dict()):
        self.data_root=data_root
        self.queues=queues
        self.data_node=data_node
        self.Xdata_node=Xdata_node
        return

    def initialize_retrieval(self):
        #Initialize variables:
        self.retrievable_vars=[var for var in self.data_root.variables.keys() 
                            if  var in self.data_root.groups['soft_links'].variables.keys()]

        #Get list of paths:
        self.paths_list=self.data_root.groups['soft_links'].variables['path'][:]
        self.checksums_list=self.data_root.groups['soft_links'].variables['checksum'][:]
        self.paths_id_list=self.data_root.groups['soft_links'].variables['path_id'][:]
        self.file_type_list=self.data_root.groups['soft_links'].variables['file_type'][:]
        self.version_list=self.data_root.groups['soft_links'].variables['version'][:]

        self.variables=dict()
        return

    def replicate(self,output,check_empty=False):
        #replicate attributes
        netcdf_utils.replicate_netcdf_file(output,self.data_root)
        #replicate and copy variables:
        for var_name in self.data_root.variables.keys():
            netcdf_utils.replicate_and_copy_variable(output,self.data_root,var_name,check_empty=check_empty)
        if 'soft_links' in self.data_root.groups.keys():
            output_grp=netcdf_utils.replicate_group(output,self.data_root,'soft_links')
            netcdf_utils.replicate_netcdf_file(output_grp,self.data_root.groups['soft_links'])
            for var_name in self.data_root.groups['soft_links'].variables.keys():
                netcdf_utils.replicate_and_copy_variable(output_grp,self.data_root.groups['soft_links'],var_name)
        return

    def retrieve_time_axis(self,years=None,months=None,days=None,min_year=None):
        time_axis=self.data_root.variables['time'][:]
        time_restriction=np.ones(time_axis.shape,dtype=np.bool)
        if years!=None or months!=None or days!=None:
            date_axis=netcdf_utils.get_date_axis(self.data_root.variables['time'])
            if years!=None:
                years_axis=np.array([date.year for date in date_axis])
                if min_year!=None:
                    #Important for piControl:
                    #time_restriction=np.logical_and(time_restriction,years_axis-years_axis.min()+min_year== years)
                    time_restriction=np.logical_and(time_restriction,[True if year in years else False for year in years_axis-years_axis.min()+min_year])
                else:
                    #time_restriction=np.logical_and(time_restriction,years_axis== years)
                    time_restriction=np.logical_and(time_restriction,[True if year in years else False for year in years_axis])
            if months!=None:
                months_axis=np.array([date.month for date in date_axis])
                #time_restriction=np.logical_and(time_restriction,months_axis==month)
                time_restriction=np.logical_and(time_restriction,[True if month in months else False for month in months_axis])
            if days!=None:
                days_axis=np.array([date.day for date in date_axis])
                time_restriction=np.logical_and(time_restriction,[True if month in days else False for month in days_axis])
        return time_axis,time_restriction

    def retrieve(self,output,retrieval_function,year=None,month=None,day=None,min_year=None,source_dir=None,semaphores=[]):
        self.initialize_retrieval()
        if source_dir!=None:
            #Check if the file has already been retrieved:
            self.paths_list,self.file_type_list=retrieval_utils.find_local_file(source_dir,self.data_root.groups['soft_links'])

        #Define tree:
        self.tree=self.data_root.path.split('/')[1:]

        #Then find time axis, time restriction and which variables to retrieve:
        time_axis, time_restriction=self.retrieve_time_axis(years=year,months=month,days=day,min_year=min_year)

        #Record to output if output is a netCDF4 Dataset:
        if (isinstance(output,netCDF4.Dataset) or
            isinstance(output,netCDF4.Group)):

            if not 'time' in output.dimensions.keys():
                netcdf_utils.create_time_axis(output,self.data_root,time_axis[time_restriction])

            #Replicate all the other variables:
            for var in set(self.data_root.variables.keys()).difference(self.retrievable_vars):
                if not var in output.variables.keys():
                    output=netcdf_utils.replicate_and_copy_variable(output,self.data_root,var)
                    #output=netcdf_utils.replicate_netcdf_var(output,self.data_root,var)
                    #output.variables[var][:]=self.data_root.variables[var][:]

        for var_to_retrieve in self.retrievable_vars:
            self.retrieve_variables(retrieval_function,var_to_retrieve,time_restriction,
                                        output,semaphores=semaphores)
                                        #paths_list,file_type_list,paths_id_list,checksums_list,version_list,
        return

    def open(self):
        self.initialize_retrieval()
        self.tree=[]
        self.output_root=netCDF4.Dataset('temp_file.pid'+str(os.getpid()),
                                      'w',format='NETCDF4',diskless=True,persist=False)
        return

    def assign(self,var_to_retrieve,time_restriction):
        time_axis, time_bool=self.retrieve_time_axis()

        self.output_root.createGroup(var_to_retrieve)
        netcdf_utils.create_time_axis(self.output_root.groups[var_to_retrieve],self.data_root,time_axis[np.array(time_restriction)])
        self.retrieve_variables('retrieve_path_data',var_to_retrieve,np.array(time_restriction),self.output_root.groups[var_to_retrieve])
        for var in self.output_root.groups[var_to_retrieve].variables.keys():
            self.variables[var]=self.output_root.groups[var_to_retrieve].variables[var]
        return

    def close(self):
        self.output_root.close()
        return

    def retrieve_variables(self,retrieval_function,var_to_retrieve,time_restriction,
                                            output,semaphores=None):
                                #paths_list,file_type_list,paths_id_list,checksums_list,version_list,
        paths_link=self.data_root.groups['soft_links'].variables[var_to_retrieve][time_restriction,0]
        indices_link=self.data_root.groups['soft_links'].variables[var_to_retrieve][time_restriction,1]

        #Convert paths_link to id in path dimension:
        paths_link=np.array([list(self.paths_id_list).index(path_id) for path_id in paths_link])

        #Sort the paths so that we query each only once:
        unique_paths_list_id, sorting_paths=np.unique(paths_link,return_inverse=True)
        
        #Replicate variable to output:
        if (isinstance(output,netCDF4.Dataset) or
            isinstance(output,netCDF4.Group)):
            output=netcdf_utils.replicate_netcdf_var(output,self.data_root,var_to_retrieve,chunksize=-1,zlib=True)
            #file_path=output.filepath()
            file_path=None
        else:
            file_path=output

        dimensions=dict()
        unsort_dimensions=dict()
        dims_length=[]
        for dim in self.data_root.variables[var_to_retrieve].dimensions:
            if dim != 'time':
                if dim in self.data_root.variables.keys():
                    dimensions[dim] = self.data_root.variables[dim][:]
                else:
                    dimensions[dim] = np.arange(len(self.data_root.dimensions[dim]))
                unsort_dimensions[dim] = None
                dims_length.append(len(dimensions[dim]))
        #Maximum number of time step per request:
        max_request=450 #maximum request in Mb
        max_time_steps=int(np.floor(max_request*1024*1024/(32*np.prod(dims_length))))
        for unique_path_id, path_id in enumerate(unique_paths_list_id):
            path_to_retrieve=self.paths_list[self.paths_id_list[path_id]]

            #Next, we check if the file is available. If it is not we replace it
            #with another file with the same checksum, if there is one!
            remote_data=remote_netcdf.remote_netCDF(path_to_retrieve.replace('fileServer','dodsC'),semaphores)
            path_to_retrieve=remote_data.check_if_available_and_find_alternative([path.replace('fileServer','dodsC') for path in self.paths_list],
                                                                      self.checksums_list).replace('dodsC','fileServer')
            file_type=self.file_type_list[list(self.paths_list).index(path_to_retrieve)]
            version='v'+str(self.version_list[list(self.paths_list).index(path_to_retrieve)])
            checksum=self.checksums_list[list(self.paths_list).index(path_to_retrieve)]

            #Append the checksum:
            path_to_retrieve+='|'+checksum

            #time_indices=sorted_indices_link[sorted_paths_link==path_id]
            time_indices=indices_link[sorting_paths==unique_path_id]
            num_time_chunk=int(np.ceil(len(time_indices)/float(max_time_steps)))
            for time_chunk in range(num_time_chunk):
                time_slice=slice(time_chunk*max_time_steps,(time_chunk+1)*max_time_steps,1)
                dimensions['time'], unsort_dimensions['time'] = indices_utils.prepare_indices(time_indices[time_slice])
                
                #Get the file tree:
                args = ({'path':path_to_retrieve,
                        'var':var_to_retrieve,
                        'indices':dimensions,
                        'unsort_indices':unsort_dimensions,
                        'sort_table':np.arange(len(sorting_paths))[sorting_paths==unique_path_id][time_slice],
                        'file_path':file_path,
                        'version':version},
                        copy.deepcopy(self.tree))
                        #'sort_table':np.argsort(sorting_paths)[sorted_paths_link==path_id][time_slice],

                #Retrieve only if it is from the requested data node:
                data_node=retrieval_utils.get_data_node(path_to_retrieve,file_type)
                if nc_Database.is_level_name_included_and_not_excluded('data_node',self,data_node):
                    if data_node in self.queues.keys():
                        print 'Recovering '+'/'.join(self.tree)
                        self.queues[data_node].put((retrieval_function,)+copy.deepcopy(args))
                    else:
                        if (isinstance(output,netCDF4.Dataset) or
                            isinstance(output,netCDF4.Group)):
                            netcdf_utils.assign_tree(output,*getattr(retrieval_utils,retrieval_function)(args[0],args[1]))
        return 




import netCDF4
import numpy as np
import retrieval_utils
import netcdf_utils
import time
import sys

class remote_netCDF:
    def __init__(self,netcdf_file_name,semaphores):
        self.file_name=netcdf_file_name
        self.semaphores=semaphores
        self.remote_data_node=retrieval_utils.get_data_node(self.file_name, 'HTTPServer')
        if isinstance(semaphores,dict):
            self.in_semaphores=(self.remote_data_node in  self.semaphores.keys())
        else:
            self.in_semaphores=False
        self.Dataset=None
        return

    def open(self):
        self.acquire_semaphore()
        try:
            self.Dataset=netCDF4.Dataset(self.file_name)
        except:
            self.close()
        return
    
    def close(self):
        try:
            if isinstance(self.Dataset,netCDF4.Dataset):
                self.Dataset.close()
        except:
            pass
        #del self.Dataset
        self.Dataset=None
        self.release_semaphore()
        return

    def acquire_semaphore(self):
        if self.in_semaphores:
            self.semaphores[self.remote_data_node].acquire()
        return

    def release_semaphore(self):
        if self.in_semaphores:
            self.semaphores[self.remote_data_node].release()
        return

    def open_with_error(self):
        error_statement=' '.join('''
The url {0} could not be opened. 
Copy and paste this url in a browser and try downloading the file.
If it works, you can stop the download and retry using cdb_query. If
it still does not work it is likely that your certificates are either
not available or out of date.'''.splitlines()).format(self.file_name.replace('dodsC','fileServer'))
        self.acquire_semaphore()
        try:
            try:
                self.Dataset=netCDF4.Dataset(self.file_name)
            except:
                time.sleep(30)
                self.Dataset=netCDF4.Dataset(self.file_name)
        except:
            self.close()
            raise dodsError(error_statement)
        return

    def is_available(self):
        try:
            self.open_with_error()
            self.close()
            return True
        except dodsError as e:
            self.close()
            e_mod=" This is a common error and is not fatal. It could however affect the number of datasets that are kept."
            print e.value+e_mod
            return False

    def check_if_available_and_find_alternative(self,paths_list,checksums_list):
        if not self.is_available():
            checksum=checksums_list[list(paths_list).index(self.file_name)]
            for cs_id, cs in enumerate(checksums_list):
                if cs==checksum and paths_list[cs_id]!=self.file_name:
                    remote_data=remote_netCDF(paths_list[cs_id],self.semaphores)
                    if remote_data.is_available():
                        return paths_list[cs_id]
            return self.file_name
        else:
            return self.file_name

    def retrieve_time(self):
        error_statement=' '.join('''
The url {0} could not be opened. 
Copy and paste this url in a browser and try downloading the file.
If it works, you can stop the download and retry using cdb_query. If
it still does not work it is likely that your certificates are either
not available or out of date.'''.splitlines()).format(self.file_name.replace('dodsC','fileServer'))
        try:
            try:
                if 'calendar' in self.Dataset.variables['time'].ncattrs():
                    calendar=self.Dataset.variables['time'].calendar
                else:
                    calendar='standard'
                units=self.Dataset.variables['time'].units
                native_time_axis=self.Dataset.variables['time'][:]
            except:
                time.sleep(30)
                if 'calendar' in self.Dataset.variables['time'].ncattrs():
                    calendar=self.Dataset.variables['time'].calendar
                else:
                    calendar='standard'
                units=self.Dataset.variables['time'].units
                native_time_axis=self.Dataset.variables['time'][:]
        except:
            raise dodsError(error_statement)
        if units=='day as %Y%m%d.%f':
            time_axis=np.array(map(netcdf_utils.convert_to_date_absolute,native_time_axis))
        else:
            try:
                time_axis=netCDF4.num2date(native_time_axis,units=units,calendar=calendar)
            except TypeError:
                time_axis=np.array([]) 
        return time_axis

    def get_time(self):
        time_axis=np.zeros((0,))
        try:
            #temp = sys.stdout
            #sys.stdout = NullDevice()
            self.open_with_error()
            #if '/'.join(self.file_name.split('/')[0:3]).split('.')[-1]=='de':
            #    print "opened "+self.file_name
            time_axis=self.retrieve_time()
            self.close()
            #sys.stdout=temp
        except dodsError as e:
            self.close()
            e_mod=" This is a common error and is not fatal. It could however affect the number of datasets that are kept."
            print e.value+e_mod
        return time_axis

class dodsError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)
class NullDevice():
    def write(self, s):
        pass



import os
import subprocess

import urllib2, httplib
from cookielib import CookieJar

import copy

import indices_utils
import remote_netcdf

import numpy as np

import hashlib

class HTTPSClientAuthHandler(urllib2.HTTPSHandler):
    def __init__(self, key, cert):
        urllib2.HTTPSHandler.__init__(self)
        self.key = key
        self.cert = cert

    def https_open(self, req):
        # Rather than pass in a reference to a connection class, we pass in
        # a reference to a function which, for all intents and purposes,
        # will behave as a constructor
        return self.do_open(self.getConnection, req)

    def getConnection(self, host, timeout=300):
        return httplib.HTTPSConnection(host, key_file=self.key, cert_file=self.cert)

#def check_file_availability(url_name):
#    wget_call='wget --spider --ca-directory={0} --certificate={1} --private-key={1}'.format(os.environ['X509_CERT_DIR'],os.environ['X509_USER_PROXY']).split(' ')
#    wget_call.append(url_name)
#
#    proc=subprocess.Popen(wget_call,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
#    (out, err) = proc.communicate()
#
#    status_string='HTTP request sent, awaiting response... '
#    error_codes=[ int(line.replace(status_string,'').split(' ')[0]) for line in err.splitlines() if status_string in line]
#    length_string='Length: '
#    lengths=[ int(line.replace(length_string,'').split(' ')[0]) for line in err.splitlines() if length_string in line]
#   
#    if 200 in error_codes and max(lengths)>0:
#        return True
#    else:
#        return False

def check_file_availability(url_name):
    cj = CookieJar()
    try: 
        opener = urllib2.build_opener(HTTPSClientAuthHandler(os.environ['X509_USER_PROXY'], os.environ['X509_USER_PROXY']),urllib2.HTTPCookieProcessor(cj))
        response = opener.open(url_name)

        if response.msg=='OK' and response.headers.getheaders('Content-Length')[0]:
            return True
        else:
            return False
    except:
        return False

def download_secure(url_name,dest_name):
    if check_file_availability(url_name):
        cj = CookieJar()
        opener = urllib2.build_opener(HTTPSClientAuthHandler(os.environ['X509_USER_PROXY'], os.environ['X509_USER_PROXY']),urllib2.HTTPCookieProcessor(cj))
        data = opener.open(url_name)

        meta = data.info()
        file_size = int(meta.getheaders("Content-Length")[0])
        size_string="Downloading: %s MB: %s" % (dest_name, file_size/2.0**20)
        
        directory=os.path.dirname(dest_name)
        if not os.path.exists(directory):
            os.makedirs(directory)
        dest_file = open(dest_name, 'wb')
        file_size_dl = 0
        block_sz = 8192
        while True:
            buffer = data.read(block_sz)
            if not buffer:
                break
        
            file_size_dl += len(buffer)
            dest_file.write(buffer)
            status = r"%10d  [%3.2f%%]" % (file_size_dl, file_size_dl * 100. / file_size)
            status = status + chr(8)*(len(status)+1)
            #print status,

        dest_file.close()
    else:
        size_string="URL %s is missing." % (url_name)
    return size_string
            
def md5_for_file(f, block_size=2**20):
    md5 = hashlib.md5()
    while True:
        data = f.read(block_size)
        if not data:
            break
        md5.update(data)
    return md5.hexdigest()

#def retrieve_path(path,out_destination):
def retrieve_path(in_dict,pointer_var):
    path=in_dict['path']
    out_destination=in_dict['file_path']
    version=in_dict['version']
    var=in_dict['var']
    dest_name=out_destination.replace('tree','/'.join(pointer_var[:-1]))
    dest_name=dest_name.replace('var',var)
    dest_name=dest_name.replace('version',version)

    decomposition=path.split('|')
    if not (isinstance(decomposition,list) and len(decomposition)>1):
        return

    root_path=decomposition[0]
    dest_name+=root_path.split('/')[-1]
    try: 
        md5sum=md5_for_file(open(dest_name,'r'))
    except:
        md5sum=''
    if md5sum==decomposition[1]:
        return 'File '+dest_name+' found. MD5 OK! Not retrieving.'

    size_string=download_secure(root_path,dest_name)
    try: 
        md5sum=md5_for_file(open(dest_name,'r'))
    except:
        md5sum=''
    if md5sum!=decomposition[1]:
        try:
            os.remove(dest_name)
        except:
            pass
        return size_string+'\n'+'File '+dest_name+' does not have the same MD5 checksum as published on the ESGF. Removing this file...'
    else:
        return size_string+'\n'+'Checking MD5 checksum of retrieved file... MD5 OK!'

def find_local_file(source_dir,data):
    paths_list=data.variables['path'][:]
    version_list=data.variables['version'][:]
    checksum_list=data.variables['checksum'][:]
    file_type_list=data.variables['file_type'][:]
    #THIS IS NOT DRS-SAFE:
    tree='/'.join(data.path.split('/')[1:-2])
    var=data.path.split('/')[-2]
    unique_paths_list=list(np.unique([source_dir+'/'+tree+'/v'+str(version)+'/'+var+'/'+path.split('/')[-1] for path, version in zip(paths_list,version_list)]))
    unique_checksum_list=[]
    for path in unique_paths_list:
        try:
            md5sum=md5_for_file(open(path,'r'))
        except:
            md5sum=''
        unique_checksum_list.append(md5sum)
    new_paths_list=[]
    new_file_type_list=[]
    for path_id,path in enumerate(paths_list):
        local_path=source_dir+'/'+tree+'/v'+str(version_list[path_id])+'/'+var+'/'+path.split('/')[-1]
        if unique_checksum_list[unique_paths_list.index(local_path)]==checksum_list[path_id]:
            new_paths_list.append(local_path)
            new_file_type_list.append('local_file')
        else:
            new_paths_list.append(path)
            new_file_type_list.append(file_type_list[path_id])
    return new_paths_list, new_file_type_list

def retrieve_path_data(in_dict,pointer_var):
    path=in_dict['path'].replace('fileServer','dodsC').split('|')[0]
    var=in_dict['var']
    indices=copy.copy(in_dict['indices'])
    unsort_indices=copy.copy(in_dict['unsort_indices'])
    sort_table=in_dict['sort_table']

    remote_data=remote_netcdf.remote_netCDF(path,[])
    remote_data.open_with_error()
    dimensions=remote_data.Dataset.variables[var].dimensions
    for dim in dimensions:
        if dim != 'time':
            if dim in remote_data.Dataset.variables.keys():
                remote_dim = remote_data.Dataset.variables[dim][:]
            else:
                remote_dim = np.arange(len(remote_data.Dataset.dimensions[dim]))
            indices[dim], unsort_indices[dim] = indices_utils.prepare_indices(
                                                            indices_utils.get_indices_from_dim(remote_dim,indices[dim]))
        
    retrieved_data=grab_remote_indices(remote_data.Dataset.variables[var],indices,unsort_indices)
    remote_data.close()
    return (retrieved_data, sort_table,pointer_var+[var])

def get_data_node(path,file_type):
    if file_type=='HTTPServer':
        return '/'.join(path.split('/')[:3])
    elif file_type=='local_file':
        return '/'.join(path.split('/')[:2])
    else:
        return ''


#def add_axis(array,axis_id):
#    return np.reshape(array,(1,)+array.shape)

def grab_remote_indices(variable,indices,unsort_indices):
    dimensions=variable.dimensions
    return retrieve_slice(variable,indices,unsort_indices,dimensions[0],dimensions[1:],0)

def retrieve_slice(variable,indices,unsort_indices,dim,dimensions,dim_id,getitem_tuple=tuple()):
    if len(dimensions)>0:
        return np.take(np.concatenate(map(lambda x: retrieve_slice(variable,
                                                 indices,
                                                 unsort_indices,
                                                 dimensions[0],
                                                 dimensions[1:],
                                                 dim_id+1,
                                                 getitem_tuple=getitem_tuple+(x,)),
                                                 indices[dim]),
                              axis=dim_id),unsort_indices[dim],axis=dim_id)
    else:
        return np.take(np.concatenate(map(lambda x: variable.__getitem__(getitem_tuple+(x,)),
                                                 indices[dim]),
                              axis=dim_id),unsort_indices[dim],axis=dim_id)



import sqlalchemy
import sqlalchemy.orm
import json

import os

import retrieval_utils
import netcdf_soft_links
import netCDF4
import netcdf_utils

import copy

class nc_Database:
    def __init__(self,project_drs,database_file=None):
        #Defines the tree structure:
        self.drs=project_drs
        self.database_file=database_file

        self._setup_database()

        return

    def _setup_database(self):
        #Create an in-memory sqlite database, for easy subselecting.
        #Uses sqlalchemy
        self.engine = sqlalchemy.create_engine('sqlite:///:memory:', echo=False)
        self.metadata = sqlalchemy.MetaData(bind=self.engine)

        self.time_db = sqlalchemy.Table('time_db',self.metadata,
                sqlalchemy.Column('case_id',sqlalchemy.Integer,primary_key=True),
                *(sqlalchemy.Column(level_name, sqlalchemy.String(255)) for level_name in self.drs.base_drs)
                )
        self.metadata.create_all()
        sqlalchemy.orm.clear_mappers()
        sqlalchemy.orm.mapper(File_Expt,self.time_db)

        self.file_expt = File_Expt(self.drs.base_drs)
        self.session = sqlalchemy.orm.create_session(bind=self.engine, autocommit=False, autoflush=True)
        return

    def close_database(self):
        self.session.close()
        self.engine.dispose()
        self._setup_database()
        return

    def load_nc_file(self):
        self.Dataset=netCDF4.Dataset(self.database_file,'r')
        return

    def close_nc_file(self):
        self.Dataset.close()
        del self.Dataset
        return

    def load_header(self):
        self.load_nc_file()
        #Load header:
        header=dict()
        for att in set(self.drs.header_desc).intersection(self.Dataset.ncattrs()):
            header[att]=json.loads(self.Dataset.getncattr(att))
        self.close_nc_file()
        return header

    def record_header(self,output_root,header):
        for value in header.keys():
            output_root.setncattr(value,json.dumps(header[value]))
        return

    def populate_database(self,options,find_function):
        self.load_nc_file()
        self.file_expt.time='0'
        populate_database_recursive(self,self.Dataset,options,find_function)
        self.close_nc_file()
        return

    def simulations_list(self):
        subset_desc=(getattr(File_Expt,item) for item in self.drs.simulations_desc)
        simulations_list=self.list_subset(subset_desc)
        return simulations_list

    def list_subset(self,subset):
        subset_list=self.session.query(*subset).distinct().all()
        return subset_list

    def list_fields(self,fields_to_list):
        fields_list=sorted(list(set(self.list_subset((getattr(File_Expt,field) for field in fields_to_list)))))
        return fields_list

    def list_data_nodes(self,options):
        data_node_list=self.list_subset((File_Expt.data_node,))
        return  [data_node[0] for data_node in data_node_list  
                if is_level_name_included_and_not_excluded('data_node',options,data_node)]
                
    def list_paths_by_data_node(self,data_node):
        return self.session.query(File_Expt.path).filter(File_Expt.data_node==data_node).first()[0]

    def list_paths(self):
        #subset=tuple([File_Expt.path,File_Expt.file_type]+[getattr(File_Expt,item) for item in self.drs.official_drs])
        subset=tuple([File_Expt.path,]+[getattr(File_Expt,item) for item in self.drs.official_drs])
        return sorted(list(set(self.list_subset(subset))))

    def write_database(self,header,options,record_function_handle,semaphores=[]):
        #List all the trees:
        drs_list=copy.copy(self.drs.base_drs)

        #drs_to_remove=['search','path','file_type','version','time']
        drs_to_remove=['path','data_node','file_type','version','time']
        for drs in drs_to_remove: drs_list.remove(drs)
        #Remove the time:
        drs_to_remove.remove('time')

        #Find the unique tuples:
        trees_list=self.list_subset([getattr(File_Expt,level) for level in drs_list])

        #Create output:
        filepath=options.out_diagnostic_netcdf_file+'.pid'+str(os.getpid())
        output_root=netCDF4.Dataset(filepath,
                                      'w',format='NETCDF4',diskless=True,persist=True)
        #output_root=netCDF4.Dataset(options.out_diagnostic_netcdf_file+'.pid'+str(os.getpid()),
        #                              'w',format='NETCDF4')
        netcdf_pointers=netcdf_soft_links.create_netCDF_pointers(
                                                          header['file_type_list'],
                                                          header['data_node_list'],
                                                          semaphores=semaphores)
        self.record_header(output_root,header)
        temp_string=''
        for att in self.drs.simulations_desc:
            if ( att in dir(options) and
                 getattr(options,att)!=None):
                temp_string+=att+': '+str(getattr(options,att))
        if len(temp_string)>0:
            output_root.setncattr('cdb_query_temp',temp_string)
                    

        #Define time subset:
        if 'months_list' in header.keys():
            months=header['months_list']
        else:
            months=range(1,13)
        
        for tree in trees_list:
            time_frequency=tree[drs_list.index('time_frequency')]
            experiment=tree[drs_list.index('experiment')]
            var=tree[drs_list.index('var')]
            conditions=[ getattr(File_Expt,level)==value for level,value in zip(drs_list,tree)]
            out_tuples=[ getattr(File_Expt,level) for level in drs_to_remove]
            #Find list of paths:
            paths_list=[{drs_name:path[drs_id] for drs_id, drs_name in enumerate(drs_to_remove)} for path in self.session.query(*out_tuples
                                    ).filter(sqlalchemy.and_(*conditions)).distinct().all()]

            output=create_tree(output_root,zip(drs_list,tree))
            #Record data:
            if (time_frequency in ['fx','clim'] and record_function_handle!='record_paths'):
                netcdf_pointers.record_fx(output,paths_list,var)
            else:
                years=[ int(year) for year in header['experiment_list'][experiment].split(',')]
                getattr(netcdf_pointers,record_function_handle)(output,
                                                                paths_list,
                                                                var,years,months)

            #Remove recorded data from database:
            self.session.query(*out_tuples).filter(sqlalchemy.and_(*conditions)).delete()

        return output_root, filepath

    def retrieve_database(self,options,output,queues,retrieval_function):
        self.load_nc_file()
        retrieve_tree_recursive(options,self.Dataset,output,queues,retrieval_function)
        if 'ensemble' in dir(options) and options.ensemble!=None:
            #Always include r0i0p0 when ensemble was sliced:
            options_copy=copy.copy(options)
            options_copy.ensemble='r0i0p0'
            retrieve_tree_recursive(options_copy,self.Dataset,output,queues,retrieval_function)
        self.close_nc_file()
        return


#####################################################################
#####################################################################
#########################  DATABASE CONVERSION
#####################################################################
#####################################################################

def populate_database_recursive(nc_Database,data,options,find_function):
    if 'soft_links' in data.groups.keys():
        soft_links=data.groups['soft_links']
        paths=soft_links.variables['path'][:]
        for path_id, path in enumerate(paths):
            #id_list=['file_type','search']
            id_list=['file_type']
            for id in id_list:
                setattr(nc_Database.file_expt,id,soft_links.variables[id][path_id])

            #Check if data_node was included:
            data_node=retrieval_utils.get_data_node(soft_links.variables['path'][path_id],
                                                    soft_links.variables['file_type'][path_id])

            if is_level_name_included_and_not_excluded('data_node',options,data_node):
                setattr(nc_Database.file_expt,'path','|'.join([soft_links.variables['path'][path_id],
                                                       soft_links.variables['checksum'][path_id]]))
                setattr(nc_Database.file_expt,'version','v'+str(soft_links.variables['version'][path_id]))
                setattr(nc_Database.file_expt,'data_node',data_node)
                find_function(nc_Database,copy.deepcopy(nc_Database.file_expt))
    elif len(data.groups.keys())>0:
        for group in data.groups.keys():
            level_name=data.groups[group].getncattr('level_name')
            if is_level_name_included_and_not_excluded(level_name,options,group):
                setattr(nc_Database.file_expt,data.groups[group].getncattr('level_name'),group)
                populate_database_recursive(nc_Database,data.groups[group],options,find_function)
    elif 'path' in data.ncattrs():
        #for fx variables:
        #id_list=['file_type','search']
        id_list=['file_type']
        for id in id_list:
            setattr(nc_Database.file_expt,id,data.getncattr(id))
        setattr(nc_Database.file_expt,'path','|'.join([data.getncattr('path'),
                                               data.getncattr('checksum')]))
        setattr(nc_Database.file_expt,'version',str(data.getncattr('version')))

        setattr(nc_Database.file_expt,'data_node',
                    retrieval_utils.get_data_node(nc_Database.file_expt.path,
                                                  nc_Database.file_expt.file_type))
        find_function(nc_Database,copy.deepcopy(nc_Database.file_expt))
    else:
        #for retrieved datasets:
        #id_list=['file_type','search','path','version']
        id_list=['file_type','path','version']
        for id in id_list:
            setattr(nc_Database.file_expt,id,'')
        if len(data.variables.keys())>0:
            setattr(nc_Database.file_expt,'data_node',
                        retrieval_utils.get_data_node(nc_Database.file_expt.path,
                                                      nc_Database.file_expt.file_type))
            find_function(nc_Database,copy.deepcopy(nc_Database.file_expt))
    return

def create_tree(output_root,tree):
    return create_tree_recursive(output_root,tree)

def create_tree_recursive(output_top,tree):
    level_name=tree[0][1]
    if not level_name in output_top.groups.keys(): 
        output=output_top.createGroup(level_name)
        output.level_name=tree[0][0]
    else:
        output=output_top.groups[level_name]
    if len(tree)>1:
        output=create_tree_recursive(output,tree[1:])
    return output

def retrieve_tree_recursive(options,data,output,queues,retrieval_function):
    if 'soft_links' in data.groups.keys():
        kwargs={'queues':queues}
        if 'data_node' in dir(options) and 'Xdata_node' in dir(options):
            kwargs['data_node']=options.data_node
            kwargs['Xdata_node']=options.Xdata_node
        elif 'data_node' in dir(options):
            kwargs['data_node']=options.data_node
        elif 'Xdata_node' in dir(options):
            kwargs['Xdata_node']=options.Xdata_node
            
        netcdf_pointers=netcdf_soft_links.read_netCDF_pointers(data,**kwargs)
        netcdf_pointers.retrieve(output,
                                 retrieval_function,
                                 year=options.year,
                                 month=options.month,
                                 day=options.day,
                                 min_year=options.min_year,
                                 source_dir=options.source_dir)
    elif len(data.groups.keys())>0:
        for group in data.groups.keys():
            level_name=data.groups[group].getncattr('level_name')
            if ( is_level_name_included_and_not_excluded(level_name,options,group) and
                 retrieve_tree_recursive_check_not_empty(options,data.groups[group])):
                if (isinstance(output,netCDF4.Dataset) or
                    isinstance(output,netCDF4.Group)):
                    if not group in output.groups.keys():
                        output_grp=output.createGroup(group)
                    else:
                        output_grp=output.groups[group]
                    for att in data.groups[group].ncattrs():
                        if not att in output_grp.ncattrs():
                            output_grp.setncattr(att,data.groups[group].getncattr(att))
                else:
                    output_grp=output
                retrieve_tree_recursive(options,data.groups[group],output_grp,queues,retrieval_function)
    else:
        #Fixed variables. Do not retrieve, just copy:
        if (isinstance(output,netCDF4.Dataset) or
            isinstance(output,netCDF4.Group)):
            if len(data.variables.keys())>0:
                for var in data.variables.keys():
                    #output_fx=netcdf_utils.replicate_netcdf_var(output,data,var,chunksize=-1,zlib=True)
                    #output_fx.variables[var][:]=data.variables[var][:]
                    output_fx=netcdf_utils.replicate_and_copy_variable(output,data,var,chunksize=-1,zlib=True)
                output_fx.sync()
    return

def retrieve_tree_recursive_check_not_empty(options,data):
    if 'soft_links' in data.groups.keys():
        return True
    elif len(data.groups.keys())>0:
        empty_list=[]
        for group in data.groups.keys():
            level_name=data.groups[group].getncattr('level_name')
            if is_level_name_included_and_not_excluded(level_name,options,group):
                empty_list.append(retrieve_tree_recursive_check_not_empty(options,data.groups[group]))
        return any(empty_list)
    else:
        if len(data.variables.keys())>0:
            return True
        else:
            return False

def is_level_name_included_and_not_excluded(level_name,options,group):
    if level_name in dir(options):
        if isinstance(getattr(options,level_name),list):
            included=((getattr(options,level_name)==[]) or
                     (group in getattr(options,level_name)))
        else:
            included=((getattr(options,level_name)==None) or 
                       (getattr(options,level_name)==group)) 
    else:
        included=True

    if 'X'+level_name in dir(options):
        if isinstance(getattr(options,'X'+level_name),list):
            not_excluded=((getattr(options,'X'+level_name)==[]) or
                     (not group in getattr(options,'X'+level_name)))
        else:
            not_excluded=((getattr(options,'X'+level_name)==None) or 
                           (getattr(options,'X'+level_name)!=group)) 
    else:
        not_excluded=True
    return included and not_excluded

def record_to_file(output_root,output):
    netcdf_utils.replicate_netcdf_file(output_root,output)
    netcdf_utils.replicate_full_netcdf_recursive(output_root,output,check_empty=True)
    return

class File_Expt(object):
    #Create a class that can be used with sqlachemy:
    def __init__(self,diag_tree_desc):
        for tree_desc in diag_tree_desc:
            setattr(self,tree_desc,'')



import os
import glob
import copy
import retrieval_utils
import nc_Database

def get_immediate_subdirectories(path):
    return [name for name in os.listdir(path)
                if os.path.isdir(os.path.join(path, name))]

def descend_tree(database,search_path,options,list_level=None):
    filesystem_file_type='local_file'
    only_list=[]
    expanded_search_path=os.path.abspath(os.path.expanduser(os.path.expandvars(search_path)))
    if filesystem_file_type in database.header['file_type_list']:
        description={
                   'file_type':filesystem_file_type,
                   'data_node':retrieval_utils.get_data_node(expanded_search_path,filesystem_file_type),
                   'time':'0'}
        file_expt_copy=copy.deepcopy(database.nc_Database.file_expt)
        for att in description.keys():
            setattr(file_expt_copy,att,description[att])

        only_list.append(descend_tree_recursive(database,file_expt_copy,
                                [item for item in database.drs.base_drs if not item in description.keys()],
                                expanded_search_path,
                                options,list_level=list_level))

        if 'alt_base_drs' in dir(database.drs):
            only_list.append(descend_tree_recursive(database,file_expt_copy,
                                    [item for item in database.drs.alt_base_drs if not item in description.keys()],
                                    expanded_search_path,
                                    options,list_level=list_level,alt=True))
    return [item for sublist in only_list for item in sublist]
    #if len(only_list)>0:
    #    return set(only_list[0]).intersection(*only_list)
    #else:
    #    return []

def descend_tree_recursive(database,file_expt,tree_desc,top_path,options,list_level=None,alt=False):
    if not isinstance(tree_desc,list):
        return

    if len(tree_desc)==1:
        file_list=glob.glob(top_path+'/*.nc') 
        if len(file_list)>0:
            for file in file_list:
                file_expt_copy=copy.deepcopy(file_expt)
                #file_expt_copy.path='|'.join([file,retrieval_utils.md5_for_file(open(file,'r'))])
                file_expt_copy.path=file+'|'
                if alt: 
                    file_expt_copy.model_version=file_expt_copy.model.split('-')[1]
                    file_expt_copy.model='-'.join([file_expt_copy.institute,file_expt_copy.model.split('-')[0]])
                database.nc_Database.session.add(file_expt_copy)
                database.nc_Database.session.commit()
        return file_list

    local_tree_desc=tree_desc[0]
    next_tree_desc=tree_desc[1:]


    subdir_list=[]
    #Loop through subdirectories:
    for subdir in get_immediate_subdirectories(top_path):
        if local_tree_desc+'_list' in database.header_simple.keys():
            #We keep only the subdirectories that were requested
            if subdir in database.header_simple[local_tree_desc+'_list']:
                subdir_list.append(subdir)
        else:
            #We also keep the subdirectories if they represent versions
            if not (local_tree_desc=='version' and subdir=='latest'):
                subdir_list.append(subdir)

    if list_level!=None and local_tree_desc==list_level:
        return subdir_list
    else:
        only_list=[]
        for subdir in subdir_list:
            file_expt_copy=copy.deepcopy(file_expt)
            setattr(file_expt_copy,local_tree_desc,subdir)
            #Include only subdirectories that were specified if this level was specified:
            if nc_Database.is_level_name_included_and_not_excluded(local_tree_desc,options,subdir):
                only_list.append(descend_tree_recursive(database,file_expt_copy,
                                            next_tree_desc,top_path+'/'+subdir,
                                            options,list_level=list_level,alt=alt))
        return [item for sublist in only_list for item in sublist]



import sys
import os
import shutil
import argparse 
import textwrap
import json
import json_tools
import cdb_query_archive
import cdb_query_archive_class

class Open_With_Indent:
    """This class creates an open file were indentation is tracked"""

    def __init__(self,file_name,read_method):
        self.open = open(file_name,read_method)
        self._current_indent=0
    
    def writei(self,write_string):
        self.open.write(write_string.rjust(len(write_string)+self._current_indent))

    def inc_indent(self):
        self._current_indent+=3

    def dec_indent(self):
        if self._current_indent > 3:
            self._current_indent-=3
        else:
            self._current_indent=0

class Experiment_Setup:
    """ Sets an experiment for the CMIP5 archive """

    def __init__(self,options):
        for opt in dir(options):
            if opt[0] != '_' and opt not in ['ensure_value','read_file','read_module']:
                setattr(self,opt,getattr(options,opt))
        if 'years' in dir(self):
            self.years=self.years.split(',')

        desc_list=[self.diagnostic,self.center,self.model,self.rip,self.experiment,'-'.join(self.years)]
        if self.month: desc_list.append(str(self.month))
        self.description='_'.join(desc_list)
        self.runscript_file=self.runscripts_dir+'/'+self.description

    def prepare_scripts(self):
        """ Prepares the scripts for bash launch """
    
        out=Open_With_Indent(self.runscript_file,'w')

        out.writei('#!/bin/bash\n')
        if self.pbs_expt:
            #If the script is expected to be submitted to a PBS queue, output the required headers:
	    if self.queue != None: out.writei('#PBS -q {0}\n'.format(self.queue))
            out.writei('#PBS -l nodes=1:ppn={0},walltime={1}\n'.format(max(self.m_async,self.ppn),self.walltime))
            #out.writei('#PBS -l nodes=1:ppn={0}\n'.format(self.m_async))
            out.writei('#PBS -N {0}\n'.format(self.description))
            out.writei('#PBS -o {1}/pbs_out/{0}\n'.format(self.description,self.output_dir))
            out.writei('#PBS -e {1}/pbs_err/{0}\n'.format(self.description,self.output_dir))

        #Put the header to the diagnostic:
        out.writei('\n')
        out.writei('#The next variable should be empty. If your script completed and some\n')
        out.writei('#years did not process properly (usually due to a timeout on large\n')
        out.writei('#storage systems), you can list the years (space separated)\n')
        out.writei('#to recompute these years. Simply rerun or resubmit the scripts\n')
        out.writei('CDB_YEARS_FIX_LIST=""\n'.format(self.years[0],self.years[1]))
        out.writei('\n')
        out.writei('CDB_YEARS="{0},{1}"\n'.format(self.years[0],self.years[1]))
        out.writei('CDB_YEAR_START=$(echo $CDB_YEARS | awk -F\',\' \'{print $1}\')\n')
        out.writei('CDB_YEAR_END=$(echo $CDB_YEARS | awk -F\',\' \'{print $2}\')\n')
        out.writei('export CDB_MODEL="{0}"\n'.format(self.model))
        out.writei('export CDB_CENTER="{0}"\n'.format(self.center))
        out.writei('export CDB_RUN_ID="{0}"\n'.format(self.rip))
        out.writei('export CDB_EXPT="{0}"\n'.format(self.experiment))
        out.writei('export CDB_DIAG_NAME="{0}"\n'.format(self.diagnostic))
        out.writei('export CDB_DIAG_HEADER="{0}"\n'.format(os.path.abspath(self.in_diagnostic_headers_file)))
        out.writei('export CDB_OUT_FILE=${CDB_DIAG_NAME}_${CDB_MODEL}_${CDB_EXPT}_$(printf "%04d" $CDB_YEAR_START)_$(printf "%04d" $CDB_YEAR_END)_${CDB_RUN_ID}\n')
        out.writei('\n')
        out.writei('#SET THE OUTPUT DIRECTORY\n')
        out.writei('export CDB_OUT_DIR="{0}"\n'.format(self.output_dir))
        out.writei('mkdir -p ${CDB_OUT_DIR}\n')
        out.writei('\n')
        out.writei('#SET THE TEMP DIRECTORY\n')
        out.writei('CDB_ROOT_TEMP_DIR="{0}"\n'.format(self.temp_dir))
        out.writei('if [ -z "${PBS_JOBID}" ]; then\n')
        out.inc_indent()
        out.writei('export CDB_TEMP_DIR="${CDB_ROOT_TEMP_DIR}/${CDB_DIAG_NAME}.job_pid${$}"\n')
        out.dec_indent()
        out.writei('else\n')
        out.inc_indent()
        out.writei('export CDB_TEMP_DIR="${CDB_ROOT_TEMP_DIR}/${CDB_DIAG_NAME}.pbs_job_${PBS_JOBID}"\n')
        out.dec_indent()
        out.writei('fi\n')
        out.writei('mkdir -p ${CDB_TEMP_DIR}\n')
        out.writei('\n')
        out.writei('#Subset the data pointers file and put it in temporary directory for optimal retrieval:\n')
        out.writei('CDB_DIAG_HEADER_BASE=$(basename ${CDB_DIAG_HEADER%.gz})\n')
        out.writei('cp ${CDB_DIAG_HEADER%.gz}.nc ${CDB_TEMP_DIR}/${CDB_DIAG_HEADER_BASE}.nc\n')
        out.writei('cdb_query_archive slice --experiment=${CDB_EXPT} --model=${CDB_MODEL} --center=${CDB_CENTER} ${CDB_DIAG_HEADER} ${CDB_TEMP_DIR}/${CDB_DIAG_HEADER_BASE}\n')
        out.writei('CDB_DIAG_HEADER="${CDB_TEMP_DIR}/${CDB_DIAG_HEADER_BASE}"\n')
        out.writei('\n')
        out.writei('#This variable sets the number of timesteps to add after and before a month:\n')
        out.writei('export CDB_OFFSET=0\n')
        out.writei('\n')
        out.writei('#Get the time axis for all the variables:\n')
        for var_name in self.variable_list.keys():
            description_list=['${CDB_CENTER}','${CDB_MODEL}','${CDB_EXPT}']
            description_list.extend(self.variable_list[var_name])
            if self.variable_list[var_name][0]!='fx':
                description_list.append('${CDB_RUN_ID}')
                description_list.append(var_name)
                out_time_file='${CDB_DIAG_HEADER}.'+var_name+'_'+'_'.join(self.variable_list[var_name])+'.time.nc'
                out.writei('ncks -3 -G : -g '+'/'.join(description_list)+' -v time ${CDB_DIAG_HEADER}.nc '+out_time_file+'\n')
                out.writei('ncap2 -O -s \'empty[$time]=0.0;\' '+' '.join([out_time_file,out_time_file])+'\n')
        out.writei('if [ "$CDB_YEAR_START" -lt "10" ]; then\n')
        out.inc_indent()
        out_time_file_list=[ '${CDB_DIAG_HEADER}.'+var_name+'_'+'_'.join(self.variable_list[var_name])+'.time.nc' for 
                                var_name in self.variable_list.keys() if self.variable_list[var_name][0]!='fx' ]
        out.writei('CDB_FILE_TIME_LIST="'+' '.join(out_time_file_list)+'"\n')
        out.writei('NTIME_MIN=$(echo $(for FILE in ${CDB_FILE_TIME_LIST}; do cdo -s showyear $FILE; done| sort)| awk -F\' \' \'{print $1}\')\n') 
        out.writei('CDB_YEAR_START=$((CDB_YEAR_START+NTIME_MIN))\n')
        out.writei('CDB_YEAR_END=$((CDB_YEAR_END+NTIME_MIN))\n')
        out.dec_indent()
        out.writei('fi\n')
        out.writei('#Select the requested years:\n')
        for var_name in self.variable_list.keys():
            if self.variable_list[var_name][0]!='fx':
                out_time_file='${CDB_DIAG_HEADER}.'+var_name+'_'+'_'.join(self.variable_list[var_name])+'.time.nc'
                out.writei('cdo -selyear,$(seq -s \',\' ${CDB_YEAR_START} ${CDB_YEAR_END}) '+' '.join([out_time_file,out_time_file+'.tmp'])+'\n')
                out.writei('mv '+' '.join([out_time_file+'.tmp',out_time_file])+'\n')

        #Load script file:
        script_file=open(self.diagnostic_dir+'/'+self.diagnostic+'.sh','r')

        #Define instructions:
        instructions={
                      '#!START MONTH LOOP' : start_monthly_loop,
                      '#!END MONTH LOOP' : end_monthly_loop
                    }

        #Loop through lines:
        for line in script_file:
                out.writei(line)
                for inst in instructions.keys():
                    if line.lstrip().upper()[:len(inst)]==inst:
                        self=instructions[inst](self,out,line)

        out.writei('\n')
        out.writei('rm -rf ${CDB_TEMP_DIR}\n')
        out.writei('\n')

        #Finally add BASH code at the end of the script to structure the output like the CMIP5 DRS:
        structure_out_with_cmip5_drs(self,out)
        out.open.close()

def ramdisk_protection(self,out):
    out.writei('function cleanup_ramdisk {\n')
    out.inc_indent()
    out.writei('echo -n "Cleaning up ramdisk directory {0} on "\n'.format(self.temp_dir))
    out.writei('date\n')
    out.writei('rm -rf {0}\n'.format(self.temp_dir))
    out.writei('echo -n "done at "\n')
    out.writei('date\n')
    out.dec_indent()
    out.writei('}')

    out.writei('function trap_term {\n')
    out.inc_indent()
    out.writei('echo -n "Trapped term (soft kill) signal on "\n')
    out.writei('date\n')
    out.writei('cleanup_ramdisk\n')
    out.writei('exit\n')
    out.dec_indent()
    out.writei('}\n')

    out.writei('#trap the termination signal, and call the function trap_term when\n')
    out.writei('# that happens, so results may be saved.\n')
    out.writei('trap "trap_term" TERM\n')
    return self

def structure_out_with_cmip5_drs(self,out):
    #code to structure the output with the CMIP5 DRS:
    out.writei('#THE LAST PART OF THIS SCRIPT REORGANIZES THE OUTPUT TO CONFORM WITH THE CMIP5 DRS:\n')
    out.writei('if [ ! -z "$CDB_CMIP5_COMP_LIST" ]; then\n')
    out.inc_indent()
    out.writei('for FILE_DESC in $CDB_CMIP5_COMP_LIST; do\n')
    out.inc_indent()
    out.writei('FILE=$(echo $FILE_DESC | awk -F\';\' \'{print $1}\')\n')
    out.writei('REALM_FREQ_INFO=$(echo $FILE_DESC | awk -F\';\' \'{print $2}\')\n')
    out.writei('CDB_INFO="This output was created using the Climate Diagnostics Benchmark toolbox,\n\
                developed by F. Laliberte and P. J. Kushner, 2012, at the University of Toronto.\n\
                This code was developed as part of the ExArch project."\n')
    out.writei('\n')
    out.writei('CDB_VAR_LIST=$(cdo showname ${FILE})\n')
    out.writei('\n')
    out.writei('CDB_FREQ=`echo ${REALM_FREQ_INFO} | tr \',\' \' \' | awk -F\' \' \'{ print $1}\'`\n')
    out.writei('CDB_REALM=`echo ${REALM_FREQ_INFO} | tr \',\' \' \' | awk -F\' \' \'{ print $2}\'`\n')
    out.writei('CDB_MIP=`echo ${REALM_FREQ_INFO} | tr \',\' \' \' | awk -F\' \' \'{ print $3}\'`\n')
    out.writei('\n')
    out.writei('for CDB_VAR_NAME in $CDB_VAR_LIST; do\n')
    out.inc_indent()
    out.writei('CDB_LOCAL_RUN_ID={0}\n'.format(self.rip))
    out.writei('if [ "$CDB_FREQ" == "fx" ]; then\n')
    out.inc_indent()
    out.writei('if [ "$CDB_MIP" == "fx" ]; then\n')
    out.inc_indent()
    out.writei('CDB_LOCAL_RUN_ID="r0i0p0"\n')
    out.dec_indent()
    out.writei('fi\n')
    out.dec_indent()
    out.writei('fi\n')
    out.writei('BASE_PATH="$CDB_OUT_DIR/in/{0}/{1}/{2}/$CDB_FREQ/$CDB_REALM/$CDB_MIP/$CDB_LOCAL_RUN_ID/v$(date +%F | tr -d \'-\')/$CDB_VAR_NAME"\n'.format(self.center,self.model,self.experiment))
    out.writei('mkdir -p $BASE_PATH\n')
    out.writei('FILE_NAME="$BASE_PATH/${CDB_VAR_NAME}_${CDB_MIP}_'+self.model+'_'+self.experiment+'_${CDB_LOCAL_RUN_ID}"\n')
    out.writei('case $CDB_FREQ in\n')
    out.inc_indent()
    out.writei('fx) FILE_NAME="${FILE_NAME}";;\n')
    out.writei('clim) FILE_NAME="${FILE_NAME}";;\n')
    out.writei('3hr) FILE_NAME="${FILE_NAME}_${CDB_YEAR_START}01010000-${CDB_YEAR_END}12312100";;\n')
    out.writei('6hr) FILE_NAME="${FILE_NAME}_${CDB_YEAR_START}01010000-${CDB_YEAR_END}12311800";;\n')
    out.writei('day) FILE_NAME="${FILE_NAME}_${CDB_YEAR_START}0101-${CDB_YEAR_END}1231";;\n')
    out.writei('mon) FILE_NAME="${FILE_NAME}_${CDB_YEAR_START}01-${CDB_YEAR_END}12";;\n')
    out.writei('yr) FILE_NAME="${FILE_NAME}_${CDB_YEAR_START}-${CDB_YEAR_END}";;\n')
    out.writei('*) echo "Unrecognized output frequency $CDB_FREQ";;\n')
    out.dec_indent()
    out.writei('esac\n')
    out.writei('cdo -s -f nc4 -z zip -selname,${CDB_VAR_NAME} $FILE ${FILE_NAME}.nc\n')
    #out.writei('echo $CDB_INFO > ${FILE_NAME}.info\n')
    #out.writei('cdo -s -f nc4 -z zip -setgatt,CDB_info,${FILE_NAME}.info -selname,${CDB_VAR_NAME} $FILE ${FILE_NAME}.nc\n')
    #out.writei('rm ${FILE_NAME}.info\n')
    out.dec_indent()
    out.writei('done\n')
    out.writei('rm $FILE\n')
    out.dec_indent()
    out.writei('done\n')
    out.dec_indent()
    out.writei('fi\n')

def start_monthly_loop(self,out,line):
    #This function creates a BASH function within the script. The function processes one month.
    #Because of the monthly processing this can be parallelized.
    out.writei('function monthly_processing()\n')
    out.writei('{\n')
    out.writei('cat > $3 <<EndOfScriptHeader\n')
    out.writei('#!/bin/bash\n')
    out.writei('#This file launches a monthly diagnostic\n')
    out.writei('export CDB_YEAR=$1\n')
    out.writei('export CDB_MONTH=$2\n')
    out.writei('\n')
    out.writei('export CDB_YEAR_START=${CDB_YEAR_START}\n')
    out.writei('export CDB_YEAR_END=${CDB_YEAR_END}\n')
    out.writei('export CDB_CENTER=${CDB_CENTER}\n')
    out.writei('export CDB_MODEL=${CDB_MODEL}\n')
    out.writei('export CDB_RUN_ID=${CDB_RUN_ID}\n')
    out.writei('export CDB_EXPT=${CDB_EXPT}\n')
    out.writei('export CDB_DIAG_NAME=${CDB_DIAG_NAME}\n')
    out.writei('export CDB_DIAG_HEADER=${CDB_DIAG_HEADER}\n')
    out.writei('export CDB_TEMP_DIR=${CDB_TEMP_DIR}\n')
    out.writei('export CDB_OUT_FILE=${CDB_OUT_FILE}_$(printf "%04d\n" $CDB_YEAR)_${CDB_MONTH}\n')
    #for type_var in cdb_environment_variables.keys():
    #    for name_var in cdb_environment_variables[type_var]:
    #        out.writei('export CDB_{0}_{1}=$CDB_{0}_{1}\n'.format(name_var.upper(),type_var.upper()))
    out.writei('\n')
    out.writei('export CDB_VAR_LIST="${CDB_VAR_LIST}"\n')

    out.open.write('EndOfScriptHeader\n')
    out.writei('\n')
    out.writei('cat >> $3 <<\'EndOfScriptMain\'\n')
    out.writei('\n')

    out.writei('#GENERATE A TIME AXIS FOR THE MONTH FOR EACH VARIABLE:\n')
    out.writei('\n')
    for var_name in self.variable_list.keys():
        if self.variable_list[var_name][0]!='fx':
            in_time_file='${CDB_DIAG_HEADER}.'+var_name+'_'+'_'.join(self.variable_list[var_name])+'.time.nc'
            out_time_file='${CDB_TEMP_DIR}/${CDB_OUT_FILE}.'+var_name+'_'+'_'.join(self.variable_list[var_name])+'.time.nc'
            out.writei('if [ "$CDB_OFFSET" -gt "0" ]; then\n')
            out.writei('cdo -selsmon,${CDB_MONTH},$CDB_OFFSET -selyear,${CDB_YEAR} '+' '.join([in_time_file,out_time_file])+'\n')
            out.writei('else\n')
            out.writei('cdo -selmon,${CDB_MONTH} -selyear,${CDB_YEAR} '+' '.join([in_time_file,out_time_file])+'\n')
            out.writei('fi\n')

    return self

def end_monthly_loop(self,out,line):
    #Closing the monthly loop
    for var_name in self.variable_list.keys():
        if self.variable_list[var_name][0]!='fx':
            out_time_file='${CDB_TEMP_DIR}/${CDB_OUT_FILE}.'+var_name+'_'+'_'.join(self.variable_list[var_name])+'.time.nc'
            out.writei('rm '+out_time_file+'\n')
    out.open.write('EndOfScriptMain\n')
    out.writei('}\n')

    #MONTH LOOP --- SHOULD ALLOW A PBS IMPLEMENTATION
    out.writei('\n')
    out.writei('\n')
    out.writei('CDB_YEAR=${CDB_YEAR_START}\n')

    #Create monthly_scripts
    out.writei('if [ -z "$CDB_YEARS_FIX_LIST" ];then\n')
    out.inc_indent()
    out.writei('while [ "$CDB_YEAR" -le "$CDB_YEAR_END" ]; do\n')
    out.inc_indent()
    months_list=' '.join([str(month).zfill(2) for month in self.months_list])
    out.writei('for CDB_MONTH in '+months_list+'; do\n')
    out.inc_indent()
    out.writei('monthly_processing $CDB_YEAR $CDB_MONTH ${CDB_TEMP_DIR}/script_$( printf "%04d" ${CDB_YEAR})_${CDB_MONTH}.sh\n')
    out.dec_indent()
    out.writei('done #CDB_MONTH\n')
    out.writei('let "CDB_YEAR += 1"\n')
    out.dec_indent()
    out.writei('done #CDB_YEAR\n')
    out.dec_indent()
    out.writei('else\n')
    out.inc_indent()
    out.writei('#The user has specified a subset of years to be computed.\n')
    out.writei('for CDB_YEAR in $CDB_YEARS_FIX_LIST; do\n')
    out.inc_indent()
    out.writei('for CDB_MONTH in '+months_list+'; do\n')
    out.inc_indent()
    out.writei('monthly_processing $CDB_YEAR $CDB_MONTH ${CDB_TEMP_DIR}/script_$( printf "%04d" ${CDB_YEAR})_${CDB_MONTH}.sh\n')
    out.dec_indent()
    out.writei('done #CDB_MONTH\n')
    out.dec_indent()
    out.writei('done #CDB_YEAR\n')
    out.dec_indent()
    out.writei('fi\n')

    ramdisk_protection(self,out)
    #Process the monthly_scripts
    if self.debug:
        out.writei('ls ${CDB_TEMP_DIR}/script_????_??.sh | parallel --tmpdir ${CDB_OUT_DIR} --tty "bash {}; rm {}"\n')
    else:
        out.writei('ls ${CDB_TEMP_DIR}/script_????_??.sh | parallel --tmpdir ${CDB_OUT_DIR} -k -j'+str(self.m_async)+' "bash {}; rm {}" > ${CDB_OUT_DIR}/${CDB_OUT_FILE}.out 2>&1 \n')
        #out.writei('pid=${!}\n')
        #out.writei('${CDB_SOURCE_DIR}/parallel_ramdisk.sh $pid ${CDB_TEMP_DIR}\n')
        out.writei('rm ${CDB_TEMP_DIR}/script_????_??.sh\n')
    return self

def main():

    #Option parser
    description=textwrap.dedent('''\
    This script sets up and run a diagnostic on the CMIP5 archive
    
    Created scripts might require:
    NCO 4.0.8 w/ OPeNDAP, netCDF4/HDF5
    CDO 1.5.1 w/ OPeNDAP, netCDF4/HDF5
    gnu-parallel
    ''')
    epilog="Frederic Laliberte, Paul J. Kushner, University of Toronto, 12/2012"
    usage = "%(prog)s [options] diagnostic"
    version_num='0.2'
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                            description=description,
                            usage=usage,
                            version='%(prog)s '+version_num,
                            epilog=epilog)
    parser.add_argument("in_diagnostic_headers_file")
    parser.set_defaults(drs=None)

    #Setup options
    setup_group=parser.add_argument_group("Setup","These options must be set when using this script")
    setup_group.add_argument("--debug",dest="debug",
                      default=False, action="store_true",
                      help="Debug flag. Disables option --m_async.")
    setup_group.add_argument("--run",dest="run",
                      default=False, action="store_true",
                      help="Launches the scripts.")

    #Processing Options
    proc_group=parser.add_argument_group("Processing",
                            "Use these options to set parallelized options.\n\
                             BEWARE! Asynchronous options are largely untested!")
    proc_group.add_argument("--m_async",dest="m_async",
                      default=1,
                      help="Uses the specified # of processors for the asynchronous processing of months.")
    proc_group.add_argument("--ppn",dest="ppn",
                      default=1,
                      help="Processors per nodes. Set to max(M_ASYNC,PPN).")
    proc_group.add_argument("-P","--pbs_expt",dest="pbs_expt",
                      default=False, action="store_true",
                      help="Prepare a pbs header for each model/rip/experiment")
    proc_group.add_argument("-w","--walltime",dest="walltime",
                      default="1:00:00",
                      help="pbs job walltime.")
    proc_group.add_argument("-q","--queue",dest="queue",
                      default=None,
                      help="PBS queue type")
    proc_group.add_argument("--submit",dest="submit",
                      default=False, action="store_true",
                      help="Submit scripts to the PBS queue")

    slicing_args={
                  'center': [str,'Modelling center name'],
                  'model': [str,'Model name'],
                  'experiment': [str,'Experiment name'],
                  'rip': [str,'RIP identifier, e.g. r1i1p1'],
                  'var': [str,'Variable name, e.g. tas'],
                  'frequency': [str,'Frequency, e.g. day'],
                  'realm': [str,'Realm, e.g. atmos'],
                  'cmor_table': [str,'CMOR table name, e.g. day'],
                  'year': [int,'Year'],
                  'month': [int,'Month as an integer ranging from 1 to 12'],
                  'file_type': [str,'File type: OPEnDAP, local_file, HTTPServer, GridFTP']
                  }

    #Slicing options
    slicing_group=parser.add_argument_group("Slicing",
                            "Use these options to restrict the processing to a subset of models")
    for arg in slicing_args.keys():
        slicing_group.add_argument('--'+arg,type=slicing_args[arg][0],help=slicing_args[arg][1])

    options = parser.parse_args()

    #Load diagnostic description file:
    paths_dict=cdb_query_archive_class.SimpleTree(json_tools.open_json(options),options)
    #paths_dict.pointers.slice(options)
    #paths_dict.pointers.create_database(cdb_query_archive_class.find_simple)

    simulations_list=paths_dict.simulations(options)

    diag_desc=paths_dict.header

    if options.submit: print('Submitting jobs using qsub ')

    for dir_to_set in ['output_dir','diagnostic_dir','runscripts_dir','temp_dir']:
        dir_path=os.path.expanduser(
                    os.path.expandvars(diag_desc[dir_to_set]))
        if os.path.isabs(dir_path):
            setattr(options,dir_to_set,os.path.abspath(dir_path))
        else:
            setattr(options,dir_to_set,
                    os.path.abspath(dir_path))

    options.diagnostic=diag_desc['diagnostic_name']

    options.variable_list=diag_desc['variable_list']

    #Check if only a subset of months were requested:
    if 'months_list' in diag_desc.keys():
        options.months_list=diag_desc['months_list']
    else:
        options.months_list=range(1,13)

    if options.month:
        options.months_list=[options.month]

    for simulation in paths_dict.simulations_list():
        options.center, options.model, options.rip = simulation
        for exp in diag_desc['experiment_list'].keys():
            options.experiment=exp
            period_list=diag_desc['experiment_list'][exp]
            if not isinstance(period_list,list): period_list=[period_list]
            map(lambda x:period_script(x,options),period_list)

    if not options.run and not options.submit:
        print 'Scripts are available in '+options.runscripts_dir

def period_script(period,options):
    options.years=set_period(period,options)

    #PREPARE SCRIPTS
    experiment = Experiment_Setup(options)
    experiment.prepare_scripts()

    #options.file_type='HTTPServer'
    #print cdb_query_archive.list_unique_paths(paths_dict,options)

    if options.run:
        os.system('bash '+experiment.runscript_file)
    elif options.submit:
        os.system('qsub '+experiment.runscript_file)

def set_period(period,options):
    if not options.year:
        return period
    else:
        years_range=[int(item) for item in period.split(',')]
        years_range[1]+=1
        if options.year not in range(*years_range):
            return period
        else:
            return ','.join([str(options.year),str(options.year)])

if __name__ == "__main__":
    main()



import copy

class DRS:
    def __init__(self,project):
        self.project=project
        getattr(self,project)()
        return

    def CORDEX(self):
        self.file_types=['local_file', 'HTTPServer', 'GridFTP']
        self.remote_file_types=['HTTPServer','GridFTP']
        self.required_file_types=['HTTPServer']

        self.var_specs=['time_frequency']
        #self.simulations_desc=['domain','institute','driving_model','model_version','model','ensemble']
        self.simulations_desc=['domain','institute','driving_model','model','ensemble']

        self.header_desc=['search_list','file_type_list','months_list','data_node_list','experiment_list','variable_list']

        self.filename_drs=['var','domain','driving_model','experiment','ensemble','model','version','time_frequency']
        self.official_drs=[
              'domain',
              'institute',
              'driving_model',
              'experiment',
              'ensemble',
              'model',
              'model_version',
              'time_frequency',
              'var',
              'version'
              ]
        self.official_drs_no_version=copy.copy(self.official_drs)
        self.official_drs_no_version.remove('version')
        self.base_drs=self.official_drs+[
                  'file_type',
                  'time',
                  'data_node',
                  'path'
              ]
        
        self.slicing_args={
              'domain': [str,'Modelling domain'],
              'institute': [str,'Modelling institute name'],
              'driving_model': [str,'Drinving model name'],
              'experiment': [str,'Experiment name'],
              'ensemble': [str,'RIP identifier, e.g. r1i1p1'],
              'model': [str,'Model name'],
              'time_frequency': [str,'Frequency, e.g. day'],
              'var': [str,'Variable name, e.g. tas']
              }
              #'file_type': [str,'File type: '+','.join(self.file_types)]
        self.discover_exclude_args=['experiment','var','time_frequency','year','month','file_type']

        self.remote_fields=['institute','model','model_version','ensemble','driving_model','domain','file_type','version']
        self.known_fields=['experiment','var'] + self.var_specs

        self.alt_base_drs=[
              'domain',
              'institute',
              'driving_model',
              'experiment',
              'ensemble',
              'model',
              'time_frequency',
              'var',
              'version',
              'file_type',
              'time',
              'data_node',
              'path'
              ]
        return

    def CMIP5(self):
        self.file_types=['local_file', 'HTTPServer', 'GridFTP']
        self.remote_file_types=['HTTPServer','GridFTP']
        self.required_file_types=['HTTPServer','OPENDAP']

        self.var_specs=['time_frequency','realm','cmor_table']

        self.official_drs=['institute',
                            'model',
                            'experiment',
                            'time_frequency',
                            'realm',
                            'cmor_table',
                            'ensemble',
                            'version',
                            'var']
        self.official_drs_no_version=copy.copy(self.official_drs)
        self.official_drs_no_version.remove('version')

        self.filename_drs=['var','cmor_table','model','experiment','ensemble']

        self.header_desc=['search_list','file_type_list','months_list','data_node_list','experiment_list','variable_list']
        self.base_drs=self.official_drs+[
                  'file_type',
                  'time',
                  'data_node',
                  'path'
                  ]
        self.simulations_desc=['institute','model','ensemble']

        self.slicing_args={
                      'institute': [str,'Modelling institute name'],
                      'model': [str,'Model name'],
                      'experiment': [str,'Experiment name'],
                      'var': [str,'Variable name, e.g. tas'],
                      'time_frequency': [str,'Frequency, e.g. day'],
                      'realm': [str,'Realm, e.g. atmos'],
                      'cmor_table': [str,'CMOR table name, e.g. day'],
                      'ensemble': [str,'RIP identifier, e.g. r1i1p1']
                      }
                      #'file_type': [str,'File type: '+','.join(self.file_types)]
        self.discover_exclude_args=['experiment','var','time_frequency','realm','cmor_table','year','month','file_type']

        self.remote_fields=['institute','model','ensemble','file_type','version']

        self.known_fields=['experiment','var'] + self.var_specs

        return




import discover
import optimset

import netCDF4

import nc_Database
import netcdf_soft_links

import copy
import os

import json
import timeit

import numpy as np

import multiprocessing

import random

import retrieval_utils

import datetime
import time

import netcdf_utils

import sys
from StringIO import StringIO

class SimpleTree:
    def __init__(self,options,project_drs):

        self.drs=project_drs
        return

    def union_header(self):
        #This method creates a simplified header

        #Create the diagnostic description dictionary:
        self.header_simple={}
        #Find all the requested realms, frequencies and variables:

        variable_list=['var_list']+[field+'_list' for field in self.drs.var_specs]
        for list_name in variable_list: self.header_simple[list_name]=[]
        for var_name in self.header['variable_list'].keys():
            self.header_simple['var_list'].append(var_name)
            for list_id, list_name in enumerate(list(variable_list[1:])):
                self.header_simple[list_name].append(self.header['variable_list'][var_name][list_id])

        #Find all the requested experiments and years:
        experiment_list=['experiment_list','years_list']
        for list_name in experiment_list: self.header_simple[list_name]=[]
        for experiment_name in self.header['experiment_list'].keys():
            self.header_simple['experiment_list'].append(experiment_name)
            for list_name in list(experiment_list[1:]):
                self.header_simple[list_name].append(self.header['experiment_list'][experiment_name])
                
        #Find the unique members:
        for list_name in self.header_simple.keys(): self.header_simple[list_name]=list(set(self.header_simple[list_name]))
        return

    def ask(self,options):
        #Load header:
        try:
            self.header=json.load(open(options.in_diagnostic_headers_file,'r'))['header']
        except ValueError as e:
            print 'The input diagnostic file '+options.in_diagnostic_headers_file+' does not conform to JSON standard. Make sure to check its syntax'
            raise

        #Simplify the header:
        self.union_header()

        if options.list_only_field!=None:
            for field_name in discover.discover(self,options):
                print field_name
            return
        else:
            simulations_list=discover.discover_simulations_recursive(self,options,self.drs.simulations_desc)
            print "This is a list of simulations that COULD satisfy the query:"
            for simulation in simulations_list:
                print ','.join(simulation)
            print "cdb_query will now attempt to confirm that these simulations have all the requested variables."
            print "This can take some time. Please abort if there are not enough simulations for your needs."

            if options.num_procs==1:
                filepath=discover.discover(self,options)
                try:
                    os.rename(filepath,filepath.replace('.pid'+str(os.getpid()),''))
                except OSError:
                    pass
            else:
                #Find the atomic simulations:
                simulations_list=discover.discover_simulations_recursive(self,options,self.drs.simulations_desc)

                manager=multiprocessing.Manager()
                output=distributed_recovery(discover.discover,self,options,simulations_list,manager)

                #Close dataset
                output.close()
        return

    def validate(self,options):
        self.load_header(options)
        #if options.data_nodes!=None:
        #    self.header['data_node_list']=options.data_nodes

        if not 'data_node_list' in self.header.keys():
            data_node_list, url_list, simulations_list =self.find_data_nodes_and_simulations(options)
            if len(data_node_list)>1:
                self.header['data_node_list']=self.rank_data_nodes(options,data_node_list,url_list)
            else:
                self.header['data_node_list']=data_node_list
        else:
            simulations_list=[]

        if options.num_procs==1:
            filepath=optimset.optimset(self,options)
            try:
                os.rename(filepath,filepath.replace('.pid'+str(os.getpid()),''))
            except OSError:
                pass
        else:
            #Find the atomic simulations:
            if simulations_list==[]:
                simulations_list=self.list_fields_local(options,self.drs.simulations_desc)
            for simulation in simulations_list:
                if simulation[-1]!='r0i0p0':
                    print '_'.join(simulation)

            #Randomize the list:
            import random
            random.shuffle(simulations_list)

            manager=multiprocessing.Manager()
            semaphores=dict()
            for data_node in  self.header['data_node_list']:
                semaphores[data_node]=manager.Semaphore(5)
            #semaphores=[]
            #original_stderr = sys.stderr
            #sys.stderr = NullDevice()
            output=distributed_recovery(optimset.optimset_distributed,self,options,simulations_list,manager,args=(semaphores,))
            #sys.stderr = original_stderr
            #Close datasets:
            output.close()
        return

    def download(self,options):
        output=netCDF4.Dataset(options.out_diagnostic_netcdf_file,'w')
        retrieval_function='retrieve_path_data'
        self.remote_retrieve_and_download(options,output,retrieval_function)
        return

    def download_raw(self,options):
        output=options.out_destination

        #Describe the tree pattern:
        if self.drs.official_drs.index('var')>self.drs.official_drs.index('version'):
            output+='/tree/version/var/'
        else:
            output+='/tree/var/version/'
            
        retrieval_function='retrieve_path'
        self.remote_retrieve_and_download(options,output,retrieval_function)
        return

    def remote_retrieve_and_download(self,options,output,retrieval_function):
        #Recover the database meta data:
        self.load_header(options)
        self.load_database(options,find_simple)

        #Find data node list:
        data_node_list=self.nc_Database.list_data_nodes(options)
        paths_list=self.nc_Database.list_paths()
        self.close_database()

        queues=define_queues(options,data_node_list)
        #Redefine data nodes:
        data_node_list=queues.keys()
        data_node_list.remove('end')

        #Check if years should be relative, eg for piControl:
        options.min_year=None
        for experiment in self.header['experiment_list']:
            min_year=int(self.header['experiment_list'][experiment].split(',')[0])
            if min_year<10:
                options.min_year=min_year
                print 'Using min year {0} for experiment {1}'.format(str(min_year),experiment)

        #Start the retrieval workers:
        #if not ('serial' in dir(options) and options.serial):
        #    processes=dict()
        #    for data_node in data_node_list:
        #        processes[data_node]=multiprocessing.Process(target=worker_retrieve, args=(queues[data_node], queues['end']))
        #        processes[data_node].start()

        #Find the data that needs to be recovered:
        self.define_database(options)
        self.nc_Database.retrieve_database(options,output,queues,getattr(retrieval_utils,retrieval_function))
        self.close_database()

        #Launch the retrieval/monitoring:
        launch_download_and_remote_retrieve(output,data_node_list,queues,retrieval_function,options)
        return

    def find_data_nodes_and_simulations(self,options):
        #We have to time the response of the data node.
        self.load_database(options,find_simple)
        simulations_list=self.nc_Database.list_fields(self.drs.simulations_desc)

        data_node_list=self.nc_Database.list_data_nodes(options)
        url_list=[self.nc_Database.list_paths_by_data_node(data_node).split('|')[0].replace('fileServer','dodsC')
                    for data_node in data_node_list ]
        self.close_database()
        return data_node_list,url_list, simulations_list

    def rank_data_nodes(self,options,data_node_list,url_list):
        data_node_list_timed=[]
        data_node_timing=[]
        for data_node_id, data_node in enumerate(data_node_list):
            url=url_list[data_node_id]
            print 'Querying '+url+' to measure response time of data node... '
            #Try opening a link on the data node. If it does not work put this data node at the end.
            number_of_trials=5
            try:
                import_string='import cdb_query.remote_netcdf;import time;'
                load_string='remote_data=cdb_query.remote_netcdf.remote_netCDF(\''+url+'\',[]);remote_data.is_available();time.sleep(2);'
                timing=timeit.timeit(import_string+load_string,number=number_of_trials)
                data_node_timing.append(timing)
                data_node_list_timed.append(data_node)
            except:
                pass
            print 'Done!'
        return list(np.array(data_node_list_timed)[np.argsort(data_node_timing)])+list(set(data_node_list).difference(data_node_list_timed))

    def list_fields(self,options):
        fields_list=self.list_fields_local(options,options.field)
        for field in fields_list:
            print ','.join(field)
        return

    def list_fields_local(self,options,fields_to_list):
        self.load_database(options,find_simple)
        fields_list=self.nc_Database.list_fields(fields_to_list)
        self.close_database()
        return fields_list

    def define_database(self,options):
        if 'in_diagnostic_netcdf_file' in dir(options):
            self.nc_Database=nc_Database.nc_Database(self.drs,database_file=options.in_diagnostic_netcdf_file)
        else:
            self.nc_Database=nc_Database.nc_Database(self.drs)
        return

    def load_header(self,options):
        self.define_database(options)
        self.header=self.nc_Database.load_header()
        self.close_database()
        return

    def record_header(self,options,output):
        self.define_database(options)
        self.header=self.nc_Database.load_header()
        self.nc_Database.record_header(output,self.header)
        self.close_database()
        return

    def load_database(self,options,find_function):
        self.define_database(options)
        self.nc_Database.populate_database(options,find_function)
        if 'ensemble' in dir(options) and options.ensemble!=None:
            #Always include r0i0p0 when ensemble was sliced:
            options_copy=copy.copy(options)
            options_copy.ensemble='r0i0p0'
            self.nc_Database.populate_database(options_copy,find_function)
        return

    def close_database(self):
        self.nc_Database.close_database()
        del self.nc_Database
        return

class NullDevice():
    def write(self, s):
        pass


def distributed_recovery(function_handle,database,options,simulations_list,manager,args=tuple()):

    #Open output file:
    output_file_name=options.out_diagnostic_netcdf_file
    output_root=netCDF4.Dataset(output_file_name,'w',format='NETCDF4')

    simulations_list_no_fx=[simulation for simulation in simulations_list if 
                                simulation[database.drs.simulations_desc.index('ensemble')]!='r0i0p0']

    queue_result=manager.Queue()
    #queue_output=manager.Queue()
    #Set up the discovery simulation per simulation:
    args_list=[]
    for simulation_id,simulation in enumerate(simulations_list_no_fx):
        options_copy=copy.copy(options)
        for desc_id, desc in enumerate(database.drs.simulations_desc):
            setattr(options_copy,desc,simulation[desc_id])
        args_list.append((function_handle,copy.copy(database),options_copy)+args+(queue_result,))
    
    #Span up to options.num_procs processes and each child process analyzes only one simulation
    #pool=multiprocessing.Pool(processes=options.num_procs,initializer=initializer,initargs=[queue_output],maxtasksperchild=1)
    if options.num_procs>1:
        pool=multiprocessing.Pool(processes=options.num_procs,maxtasksperchild=1)
        result=pool.map_async(worker_query,args_list,chunksize=1)
    for arg in args_list:
        if options.num_procs==1:
            result=worker_query(arg)
        filename=queue_result.get()
        source_data=netCDF4.Dataset(filename,'r')
        nc_Database.record_to_file(output_root,source_data)
        source_data.close()
        try:
            os.remove(filename)
        except OSError:
            pass
        output_root.sync()
    if options.num_procs>1:
        pool.close()
        pool.join()

    return output_root

def worker_query(tuple):
    result=tuple[0](*tuple[1:-1])
    sys.stdout.flush()
    sys.stderr.flush()
    tuple[-1].put(result)
    return

class MyStringIO(StringIO):
    def __init__(self, queue, *args, **kwargs):
        StringIO.__init__(self, *args, **kwargs)
        self.queue = queue
    def flush(self):
        self.queue.put((multiprocessing.current_process().name, self.getvalue()))
        self.truncate(0)

def initializer(queue):
     sys.stderr = sys.stdout = MyStringIO(queue)

def worker_retrieve(input, output):
    for tuple in iter(input.get, 'STOP'):
        result = tuple[0](tuple[1],tuple[2])
        output.put(result)
    output.put('STOP')
    return

def launch_download_and_remote_retrieve(output,data_node_list,queues,retrieval_function,options):
    #Second step: Process the queues:
    #print datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    start_time = datetime.datetime.now()
    print('Retrieving from data nodes:')
    queues_size=dict()
    for data_node in data_node_list:
        queues_size[data_node]=queues[data_node].qsize()
    string_to_print=['0'.zfill(len(str(queues_size[data_node])))+'/'+str(queues_size[data_node])+' paths from "'+data_node+'"' for
                        data_node in data_node_list]
    print ' | '.join(string_to_print)
    print 'Progress: '

    if 'serial' in dir(options) and options.serial:
        for data_node in data_node_list:
            queues[data_node].put('STOP')
            worker_retrieve(queues[data_node], queues['end'])
            for tuple in iter(queues['end'].get, 'STOP'):
                progress_report(retrieval_function,output,tuple,queues,queues_size,data_node_list,start_time)
        
    else:
        #for data_node in data_node_list:
        #    queues[data_node].put('STOP')
        processes=dict()
        for data_node in data_node_list:
            queues[data_node].put('STOP')
            processes[data_node]=multiprocessing.Process(target=worker_retrieve, args=(queues[data_node], queues['end']))
            processes[data_node].start()

        for data_node in data_node_list:
            for tuple in iter(queues['end'].get, 'STOP'):
                progress_report(retrieval_function,output,tuple,queues,queues_size,data_node_list,start_time)

    if retrieval_function=='retrieve_path_data':
        output.close()

    print
    print('Done!')
    #print datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    return

def progress_report(retrieval_function,output,tuple,queues,queues_size,data_node_list,start_time):
    if retrieval_function=='retrieve_path':
        #print '\t', queues['end'].get()
        print '\t', tuple
        elapsed_time = datetime.datetime.now() - start_time
        print str(elapsed_time)
    elif retrieval_function=='retrieve_path_data':
        #netcdf_utils.assign_tree(output,*queues['end'].get())
        netcdf_utils.assign_tree(output,*tuple)
        output.sync()
        string_to_print=[str(queues_size[data_node]-queues[data_node].qsize()).zfill(len(str(queues_size[data_node])))+
                         '/'+str(queues_size[data_node]) for
                            data_node in data_node_list]
        elapsed_time = datetime.datetime.now() - start_time
        print str(elapsed_time)+', '+' | '.join(string_to_print)+'\r',
        #for data_node in data_node_list:
        #    print data_node, queues[data_node].qsize()
    return
        
def find_simple(pointers,file_expt):
    #for item in dir(file_expt):
    #    if item[0]!='_':
    #        print getattr(file_expt,item)
    pointers.session.add(file_expt)
    pointers.session.commit()
    return

def define_queues(options,data_node_list):
    #from multiprocessing import Manager
    #manager=Manager()
    queues={data_node : multiprocessing.Queue() for data_node in data_node_list}
    #sem=manager.Semaphore()
    #semaphores={data_node : manager.Semaphore() for data_node in data_node_list}
    #semaphores={data_node : sem for data_node in data_node_list}
    queues['end']= multiprocessing.Queue()
    if 'source_dir' in dir(options) and options.source_dir!=None:
        queues[retrieval_utils.get_data_node(options.source_dir,'local_file')]=multiprocessing.Queue()
    return queues



import numpy as np
from itertools import groupby, count

def get_indices_from_dim(source,output):
    #This function finds which indices from source should be used and in which order:
    indices=np.arange(max(source.shape))[np.in1d(source,output)]
    try:
        return np.array([ indices[source[indices]==val][0] for val in output ])
    except IndexError:
        #The in1d might have encountered afloating point error. Make the equality fuzzy:
        #warnings.warn('Dimension matching was done to floating point tolerance for some model',UserWarning)
        indices=np.arange(max(source.shape))[np.array([np.any(np.isclose(output-item,0.0)) for item in source])]
        return np.array([ indices[np.isclose(source[indices]-val,0.0)][0] for val in output ])

def convert_indices_to_slices(indices):
    if len(indices)>1:
        steps=np.unique(np.diff(indices))
        optimal_step=steps[np.argmin(map(lambda x: len(convert_indices_to_slices_step(indices,x)),steps))]
        return convert_indices_to_slices_step(indices,optimal_step)
    else:
        return convert_indices_to_slices_step(indices,1)

def convert_indices_to_slices_step(indices,step):
    slices = []
    for key, it in groupby(enumerate(indices), lambda x: x[1] - step*x[0]):
        indices = [y for x, y in it]
        slices.append(slice(indices[0], indices[-1]+1,step))
        #if len(indices) == 1:
        #    slices.append(slice(indices[0],indices[0]+1))
        #else:
        #    slices.append(slice(indices[0], indices[-1]+1))
    return slices

def prepare_indices(indices):
    sort_indices=np.argsort(indices)
    #Sort:
    indices=indices[sort_indices]
    #provide the inverse:
    unsort_indices=np.argsort(sort_indices)

    #Finally, convert the indices to slices:
    indices=convert_indices_to_slices(indices)
    return indices, unsort_indices

def largest_hyperslab(slices_dict):
    return np.prod([max([slice_length(item) for item in slices_dict[dim]])
                for dim in slices_dict.keys()])

def slice_length(slice_item):
    return len(range(slice_item.start,slice_item.stop,slice_item.step))
        



import os
import hashlib

import filesystem_query
import esgf_query

import copy

import remote_archive

#import database_utils

import cdb_query_archive_parsers
import cdb_query_archive_class

import netcdf_utils

def main_CMIP5():
    main('CMIP5')
    return

def main_CORDEX():
    main('CORDEX')
    return

def main(project):
    import argparse 
    import textwrap

    #Option parser
    description=textwrap.dedent('''\
    This script queries a {0} archive. It can query:
    1. a local POSIX-based archive that follows the {0} DRS
    filename convention and directory structure.
    2. the ESGF {0} archive.

    '''.format(project))
    epilog='Frederic Laliberte, Paul Kushner 06/2014'
    version_num='1.0.0'
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                            description=description,
                            version='%(prog)s '+version_num,
                            epilog=epilog)

    project_drs=remote_archive.DRS(project)

    #Generate subparsers
    cdb_query_archive_parsers.generate_subparsers(parser,epilog,project_drs)

    options=parser.parse_args()

    #Slicing time is peculiar
    for time_opt in ['year','month']:
        if time_opt in dir(options) and getattr(options,time_opt):
            options.time=True

    #Load pointer file:
    #if options.command=='remote_retrieve':
    #    paths_dict=cdb_query_archive_class.SimpleTree(options,project_drs)
    #    getattr(paths_dict,options.command)(options)
    #elif options.command=='download':
    #    paths_dict=cdb_query_archive_class.SimpleTree(options,project_drs)
    #    getattr(paths_dict,options.command)(options)
    if options.command=='apply':
        netcdf_utils.apply(options,project_drs)
    elif options.command=='convert':
        netcdf_utils.convert(options,project_drs)
    elif 'in_diagnostic_headers_file' in dir(options):
        paths_dict=cdb_query_archive_class.SimpleTree(options,project_drs)
        #Run the command:
        getattr(paths_dict,options.command)(options)
    elif 'in_diagnostic_netcdf_file' in dir(options):
        paths_dict=cdb_query_archive_class.SimpleTree(options,project_drs)
        #Run the command:
        getattr(paths_dict,options.command)(options)
        
if __name__ == "__main__":
    main('CMIP5')



import os
import copy
import warnings
from pyesgf.search import SearchConnection

import multiprocessing as mproc
import retrieval_utils

def descend_tree(database,search_path,options,list_level=None):
    #Create the database:
    only_list=[]
    for experiment in database.header['experiment_list'].keys():
        for var_name in database.header['variable_list'].keys():
            only_list.append(experiment_variable_search(database.nc_Database,search_path,database.header['file_type_list'],options,
                                        experiment,var_name,database.header['variable_list'][var_name],list_level=list_level))
    return [item for sublist in only_list for item in sublist]
    #if len(only_list)>0:
    #    return set(only_list[0]).intersection(*only_list)
    #else:
    #    return []

def experiment_variable_search(nc_Database,search_path,file_type_list,options,
                                experiment,var_name,var_desc,list_level=None):

    conn = SearchConnection(search_path, distrib=options.distrib)

    #Search the ESGF:
    ctx = conn.new_context(project=nc_Database.drs.project,
                        experiment=experiment)
    #print ctx.facet_counts
    ctx=ctx.constrain(**{field:var_desc[field_id] for field_id, field in enumerate(nc_Database.drs.var_specs)})

    
    for field in nc_Database.drs.slicing_args:
        if field in dir(options) and getattr(options,field)!=None:
            ctx=ctx.constrain(**{field:getattr(options,field)})
    #if options.model:
    #    ctx=ctx.constrain(model=options.model)
    #if options.institute:
    #    ctx=ctx.constrain(institute=options.institute)
    #if options.ensemble:
    #    ctx=ctx.constrain(ensemble=options.ensemble)

    nc_Database.file_expt.experiment=experiment
    nc_Database.file_expt.var=var_name
    nc_Database.file_expt.time=0
    for field_id, field in enumerate(nc_Database.drs.var_specs):
        setattr(nc_Database.file_expt,field,var_desc[field_id])

    if list_level!=None:
        import socket
        try:
            return ctx.facet_counts[list_level].keys()
        except socket.error as e:
            print search_path+' is not responding. '+e.strerror
            return []
    else:
        file_list_remote=[]
        file_list_found=[]
        try:
            file_list_found=ctx.search(variable=var_name)
            file_list_remote=map(lambda x: get_urls(nc_Database.drs,x,file_type_list,var_name),file_list_found)
            file_list_remote=[item for sublist in file_list_remote for item in sublist]
        except:
            warnings.warn('Search path {0} is unresponsive at the moment'.format(search_path),UserWarning)
       
        map(lambda x: record_url(x,nc_Database),file_list_remote)
        return []

def record_url(remote_file_desc,nc_Database):
    nc_Database.file_expt.path=remote_file_desc['url']
    nc_Database.file_expt.data_node=retrieval_utils.get_data_node(remote_file_desc['url'],remote_file_desc['file_type'])
    if remote_file_desc['file_type'] in nc_Database.drs.remote_file_types and remote_file_desc['checksum']:
        nc_Database.file_expt.path+='|'+remote_file_desc['checksum']

    for val in nc_Database.drs.remote_fields:
        setattr(nc_Database.file_expt,val,remote_file_desc[val])

    #Convert unicode to string:
    for val in dir(nc_Database.file_expt):
        if val[0]!='_' and val!='case_id':
            setattr(nc_Database.file_expt,val,str(getattr(nc_Database.file_expt,val)))

    list_of_knowns=[ getattr(nc_Database.file_expt,field) for field in nc_Database.drs.known_fields] 
    list_of_retrieved=[ remote_file_desc[field] for field in nc_Database.drs.known_fields] 
    if remote_file_desc['version']:
        if (remote_file_desc['version'][1:]!='atest' and
            len([i for i,j in zip(list_of_knowns,list_of_retrieved) if i==j])==len(list_of_knowns)):
            nc_Database.session.add(copy.deepcopy(nc_Database.file_expt))
            nc_Database.session.commit()
    return nc_Database

def get_urls(drs,result,file_type_list,var_name):
    file_list_remote=[]
    if len(set(file_type_list).intersection(set(drs.remote_file_types)))>0:
        #If remote file types were requested
        fil_ctx = result.file_context()
        try:
            shard_name=fil_ctx.shards[0]
        except:
            shard_name=None
        try:
            fil = fil_ctx.search(variable=var_name)
            for item in fil:
                file_list_remote.extend(get_url_remote(item,file_type_list,drs))
        except:
            if shard_name==None:
                 warnings.warn('An unknown shard is unresponsive at the moment'.format(shard_name))
            else:
                warnings.warn('Shard {0} is unresponsive at the moment'.format(shard_name))
        
    return file_list_remote

def get_url_remote(item,file_type_list,drs):
    url_name=[]
    try:
        keys_list=item.urls.viewkeys()
    except:
        keys_list=[]

    for key in set(keys_list).intersection(file_type_list):
        file_info=dict()
        file_info['file_type']=key
        try:
            file_info['url']=item.urls[key][0][0]
        except:
            file_info['url']=None
        for val in drs.official_drs+['checksum']:
            try:
                if val=='var':
                    #this is a temporary fix to a poor design decision.
                    #to really fix this, will have to change names 'var' to variables.
                    file_info[val]=item.json['variable']
                elif val=='version':
                    #Version is poorly implemented... Try a fix:
                    version=item.json['instance_id'].split('.')[-3]
                    #Previous fix. Does not work for CORDEX
                    #version=item.json['id'].split('.')[9]
                    if version[0]=='v':
                        file_info[val]=version
                elif val=='model_version':
                    file_info[val]='v'+item.json['version']
                else:
                    file_info[val]=item.json[val]

                if isinstance(file_info[val],list): file_info[val]=str(file_info[val][0])
            except:
                file_info[val]=None
        if (file_info['checksum']!=None and 
            set(item.urls.keys()).issuperset(drs.required_file_types)):
        #if file_info['checksum']!=None:
            url_name.append(file_info)
    return url_name




import numpy as np
import math

import netCDF4
import datetime

import hashlib

import copy

import os

import cdb_query_archive_class

import json

import retrieval_utils

import netcdf_utils

import remote_netcdf

import indices_utils

import nc_Database

import random


class soft_links:
    def __init__(self,paths_list,file_type_list,data_node_list,semaphores=[]):
        self.sorts_list=['version','file_type_id','data_node_id','path_id']
        self.id_list=['data_node','file_type','path','checksum']
        self.file_type_list=file_type_list
        self.data_node_list=data_node_list
        self.semaphores=semaphores
        self.paths_list=paths_list
        self.paths_ordering=self.order_paths_by_preference()
        return

    def create(self,output):
        if not 'soft_links' in output.groups.keys():
            output_grp=output.createGroup('soft_links')
        else:
            output_grp=output.groups['soft_links']

        #OUTPUT TO NETCDF FILE PATHS DESCRIPTIONS:
        output_grp.createDimension('path',len(self.paths_ordering))
        for id in ['version','path_id']:
            output_grp.createVariable(id,np.uint32,('path',))[:]=self.paths_ordering[id]
        for id in self.id_list:
            temp=output_grp.createVariable(id,str,('path',))
            for file_id, file in enumerate(self.paths_ordering['path']):
                temp[file_id]=str(self.paths_ordering[id][file_id])
        return 

    def order_paths_by_preference(self):
        #FIND ORDERING:
        paths_desc=[]
        for id in self.sorts_list:
            paths_desc.append((id,np.uint32))
        for id in self.id_list:
            paths_desc.append((id,'a255'))
        paths_ordering=np.empty((len(self.paths_list),), dtype=paths_desc)
        for file_id, file in enumerate(self.paths_list):
            paths_ordering['path'][file_id]=file['path'].split('|')[0]
            paths_ordering['path_id'][file_id]=file_id
            paths_ordering['checksum'][file_id]=file['path'].split('|')[1]
            paths_ordering['version'][file_id]=np.uint32(file['version'][1:])
            paths_ordering['file_type'][file_id]=file['file_type']
            paths_ordering['data_node'][file_id]=retrieval_utils.get_data_node(file['path'],paths_ordering['file_type'][file_id])

        #Sort paths from most desired to least desired:
        #First order desiredness for least to most:
        data_node_order=copy.copy(self.data_node_list)[::-1]#list(np.unique(paths_ordering['data_node']))
        file_type_order=copy.copy(self.file_type_list)[::-1]#list(np.unique(paths_ordering['file_type']))
        for file_id, file in enumerate(self.paths_list):
            paths_ordering['data_node_id'][file_id]=data_node_order.index(paths_ordering['data_node'][file_id])
            paths_ordering['file_type_id'][file_id]=file_type_order.index(paths_ordering['file_type'][file_id])
        #'version' is implicitly from least to most

        #sort and reverse order to get from most to least:
        return np.sort(paths_ordering,order=self.sorts_list)[::-1]

    def _recover_time(self,path):
        path_name=str(path['path']).replace('fileServer','dodsC').split('|')[0]
        remote_data=remote_netcdf.remote_netCDF(path_name,self.semaphores)
        time_axis=remote_data.get_time()
        table_desc=[
                   ('paths','a255'),
                   ('indices','uint32')
                   ]
        table=np.empty(time_axis.shape, dtype=table_desc)
        if len(time_axis)>0:
            table['paths']=np.array([str(path_name) for item in time_axis])
            table['indices']=range(0,len(time_axis))
        return time_axis,table
    
    def obtain_table(self):
        self.time_axis, self.table= map(np.concatenate,
                        zip(*map(self._recover_time,np.nditer(self.paths_ordering))))
        return

    def reduce_paths_ordering(self):
        #CREATE LOOK-UP TABLE:
        self.paths_indices=np.empty(self.time_axis.shape,dtype=np.uint32)
        self.time_indices=np.empty(self.time_axis.shape,dtype=np.uint32)

        paths_list=[path for path in self.paths_ordering['path'] ]
        for path_id, path in enumerate(paths_list):
            self.paths_indices[path.replace('fileServer','dodsC')==self.table['paths']]=path_id

        #Remove paths that are not necessary over the requested time range:
        paths_id_list=list(np.unique([np.min(self.paths_indices[time==self.time_axis])  for time_id, time in enumerate(self.time_axis_unique)]))
        useful_file_name_list=[path.split('/')[-1] for path in self.paths_ordering['path'][paths_id_list] ]
        for file_id, file in enumerate(self.paths_ordering):
            if not file_id in paths_id_list:
                file_name=self.paths_ordering['path'][file_id].split('/')[-1]
                if file_name in useful_file_name_list:
                    equivalent_file_id=paths_id_list[useful_file_name_list.index(file_name)]
                    if self.paths_ordering['checksum'][file_id]==self.paths_ordering['checksum'][equivalent_file_id]:
                        paths_id_list.append(file_id)
        #Sort paths_ordering:
        self.paths_ordering=self.paths_ordering[np.sort(paths_id_list)]
        #Finally, set the path_id field to be following the indices in paths_ordering:
        self.paths_ordering['path_id']=range(len(self.paths_ordering))

        #Recompute the indices to paths:
        paths_list=[path for path in self.paths_ordering['path'] ]
        for path_id, path in enumerate(paths_list):
            self.paths_indices[path.replace('fileServer','dodsC')==self.table['paths']]=path_id
        return

    def unique_time_axis(self,data,years,months):
        units=data.variables['time'].units
        calendar=data.variables['time'].calendar
        time_axis = netCDF4.date2num(self.time_axis,units,calendar=calendar)
        time_axis_unique = np.unique(time_axis)

        time_axis_unique_date=netCDF4.num2date(time_axis_unique,units,calendar=calendar)

        #Include a filter on years: 
        time_desc={}
        years_range=range(*years)
        years_range+=[years_range[-1]+1]
        if years[0]<10:
            #This is important for piControl
            years_range=list(np.array(years_range)+np.min([date.year for date in time_axis_unique_date]))
            #min_year=np.min([date.year for date in time_axis_unique_date])

        valid_times=np.array([True  if (date.year in years_range and 
                                     date.month in months) else False for date in  time_axis_unique_date])
        self.time_axis_unique=time_axis_unique[valid_times]
        self.time_axis=time_axis
        return

    def create_variable(self,output,var,years,months):
        #Recover time axis for all files:
        self.obtain_table()

        if len(self.table['paths'])>0:
            #Open the first file and use its metadata to populate container file:
            remote_data=remote_netcdf.remote_netCDF(self.table['paths'][0],self.semaphores)
            #try:
            remote_data.open_with_error()
            netcdf_utils.replicate_netcdf_file(output,remote_data.Dataset)

            #Convert time axis to numbers and find the unique time axis:
            self.unique_time_axis(remote_data.Dataset,years,months)

            self.reduce_paths_ordering()
            #Create time axis in ouptut:
            netcdf_utils.create_time_axis(output,remote_data.Dataset,self.time_axis_unique)

            self.create(output)
            self.record_indices(output,remote_data.Dataset,var)
            output.sync()
            #except dodsError as e:
            #    e_mod=" This is an uncommon error. It is likely to be FATAL."
            #    print e.value+e_mod
            remote_data.close()
        return

    def record_indices(self,output,data,var):
        netcdf_utils.replicate_netcdf_var(output,data,var,chunksize=-1,zlib=True)
        #netcdf_utils.replicate_netcdf_var(output,data,var)

        for other_var in data.variables.keys():
            if ( (not 'time' in data.variables[other_var].dimensions) and 
                 (not other_var in output.variables.keys()) ):
                netcdf_utils.replicate_and_copy_variable(output,data,other_var)

        #CREATE LOOK-UP TABLE:
        output_grp=output.groups['soft_links']
        output_grp.createDimension('indices',2)
        indices=output_grp.createVariable('indices',np.str,('indices',))
        indices[0]='path'
        indices[1]='time'

        var_out = output_grp.createVariable(var,np.uint32,('time','indices'),zlib=False,fill_value=np.iinfo(np.uint32).max)

        #Create soft links:
        for time_id, time in enumerate(self.time_axis_unique):
            var_out[time_id,0]=np.min(self.paths_indices[time==self.time_axis])
            var_out[time_id,1]=self.table['indices'][np.logical_and(self.paths_indices==var_out[time_id,0],time==self.time_axis)][0]
        return


#def create_variable_soft_links2(data,output,var,time_axis,time_axis_unique,paths_indices,paths_list,table):
#    #CREATE LOOK-UP TABLE:
#    common_substring=long_substr(paths_list)
#    soft_link_numpy=np.dtype([(path.replace('/','&#47;'),np.byte) for path in paths_list]+[('time',np.uint16)])
#    soft_link=output.createCompoundType(soft_link_numpy,'soft_link')
#    netcdf_utils.replicate_netcdf_var(output,data,var,chunksize=-1,datatype=soft_link)
#    for var in data.variables.keys():
#        if not 'time' in data.variables[var].dimensions:
#            output=netcdf_utils.replicate_netcdf_var(output,data,var)
#            output.variables[var][:]=data.variables[var][:]
#    shape = (1,)+output.variables[var].shape[1:]
#
#    #Create soft links:
#    var_out=output.variables[var]
#    for time_id, time in enumerate(time_axis_unique):
#        var_temp=np.zeros(shape,dtype=soft_link_numpy)
#
#        path_id=np.min(paths_indices[time==time_axis])
#        var_temp[paths_list[path_id].replace('/','&#47;')]=1
#        var_temp['time']=table['indices'][np.logical_and(paths_indices==path_id,time==time_axis)][0]
#        var_out[time_id,...]=var_temp
#    return
#
#def long_substr(data):
#    #from http://stackoverflow.com/questions/2892931/longest-common-substring-from-more-than-two-strings-python
#    substr = ''
#    if len(data) > 1 and len(data[0]) > 0:
#        for i in range(len(data[0])):
#            for j in range(len(data[0])-i+1):
#                if j > len(substr) and all(data[0][i:i+j] in x for x in data):
#                    substr = data[0][i:i+j]
#    return substr






import argparse 
import textwrap
import datetime
import copy

def input_arguments_json(parser):
    parser.add_argument('in_diagnostic_headers_file',
                                 help='Diagnostic headers file (input)')
    return
def output_arguments_json(parser):
    parser.add_argument('out_diagnostic_headers_file',
                                 help='Diagnostic headers file (output)')
    return

def input_arguments(parser):
    parser.add_argument('in_diagnostic_netcdf_file',
                                 help='NETCDF Diagnostic paths file (input)')
    return

def output_arguments(parser):
    parser.add_argument('out_diagnostic_netcdf_file',
                                 help='NETCDF Diagnostic paths file (output)')
    return

def slicing_arguments(parser,project_drs,exclude_args=[],action_type='store'):
    #Define the data slicing arguments in a dictionnary:
    for arg in project_drs.slicing_args.keys():
        if not arg in exclude_args:
            parser.add_argument('--'+arg,action=action_type,
                                type=project_drs.slicing_args[arg][0],
                                help=project_drs.slicing_args[arg][1]
                                )
    return
def excluded_slicing_arguments(parser,project_drs,exclude_args=[],action_type='store'):
    #Define the data slicing arguments in a dictionnary:
    for arg in project_drs.slicing_args.keys():
        if not arg in exclude_args:
            parser.add_argument('--'+'X'+arg,action=action_type,
                                type=project_drs.slicing_args[arg][0],
                                help='Exclude '+project_drs.slicing_args[arg][1]
                                )
    return


def generate_subparsers(parser,epilog,project_drs):
    #Discover tree
    subparsers = parser.add_subparsers(help='Commands to discover available data on the archive',dest='command')
    #quick_ask(subparsers,epilog,project_drs)
    ask(subparsers,epilog,project_drs)
    list_fields(subparsers,epilog,project_drs)

    #Optimset tree
    validate(subparsers,epilog,project_drs)
    download(subparsers,epilog,project_drs)
    download_raw(subparsers,epilog,project_drs)
    apply(subparsers,epilog,project_drs)
    convert(subparsers,epilog,project_drs)
    return

#def quick_ask(subparsers,epilog,project_drs):
#    #Find data
#    epilog_ask=epilog
#    parser=subparsers.add_parser('quick_ask',
#                                           description=textwrap.dedent(
#                                                '''
#                                                 Returns a list of simulations that COULD satisfy the query.
#                                                 Note using \'ask\' might significantly reduce this list.
#                                                 
#                                                 This function should NOT require appropriate certificates
#                                                 to function properly. If it fails it is possible the servers are down.'''),
#                                           epilog=epilog_ask
#                                         )
#    input_arguments_json(parser)
#    parser.add_argument('--distrib',
#                                 default=False, action='store_true',
#                                 help='Distribute the search. Will likely result in a pointers originating from one node.')
#    proc_group = parser.add_argument_group('These arguments set threading options')
#    inc_group = parser.add_argument_group('Inclusions')
#    slicing_arguments(inc_group,project_drs)
#    return parser

def ask(subparsers,epilog,project_drs):
    #Find data
    epilog_ask=epilog
    parser=subparsers.add_parser('ask',
                                           description=textwrap.dedent(
                                                '''Returns pointers to models that have as a subset the requested experiments and variables.\n\
                                                 It is good practice to check the results with \'list_fields\' before
                                                 proceeding with \'validate\'.
                                                 The output of \'validate\' might depend on the order of the header attribute
                                                 \'data_node_list\' in the output file of \'ask\'. It is good practice to
                                                 reorder this attribute before proceeding with \'validate\'.
                                                 
                                                 Unlike \'validate\' this function should NOT require appropriate certificates
                                                 to function properly. If it fails it is possible the servers are down.'''),
                                           epilog=epilog_ask
                                         )
    input_arguments_json(parser)
    output_arguments(parser)
    parser.add_argument('--list_only_field',default=None, choices=project_drs.remote_fields,
                              help='When this option is used, the ask function prints only the specified field \n\
                                  for which published data COULD match the query. Does nothing to the output file.\n\
                                  Listing separate fields is usually much quicker than the discovery step.')
    parser.add_argument('--distrib',
                                 default=False, action='store_true',
                                 help='Distribute the search. Will likely result in a pointers originating from one node.')
    proc_group = parser.add_argument_group('These arguments set threading options')
    proc_group.add_argument('--num_procs',
                                 default=1, type=int,
                                 help='Use num_procs processes to perform the computation.')
    inc_group = parser.add_argument_group('Inclusions')
    slicing_arguments(inc_group,project_drs)
    return parser

def list_fields(subparsers,epilog,project_drs):
    #List_paths
    description=textwrap.dedent('List the unique combination of fields in the file.\n\
    For example:\n\
    1) to list the paths, use option \'-f path\'\n\
    2) to list the simulations tuples use options \'-f institute -f model -f ensemble\'\n\
         ').strip()
    parser=subparsers.add_parser('list_fields',
                                            description=description,
                                            formatter_class=argparse.RawTextHelpFormatter,
                                            epilog=epilog
                                           )
    input_arguments(parser)
    select_group = parser.add_argument_group('These arguments specify the structure of the output')
    select_group.add_argument('-f','--field',action='append', type=str, choices=project_drs.base_drs,
                                       help='List the field (or fields if repeated) found in the file' )
    #select_group.add_argument('--data_nodes',default=False,action='store_true',
    #                                   help='List the data nodes found in the file. Disables the --field options.' )

    data_node_group = parser.add_argument_group('Restrict search to specific data nodes')
    data_node_group.add_argument('--data_node',type=str,action='append',help='Consider only the specified data nodes')
    data_node_group.add_argument('--Xdata_node',type=str,action='append',help='Do not consider the specified data nodes')

    inc_group = parser.add_argument_group('Inclusions')
    slicing_arguments(inc_group,project_drs,action_type='append')
    exc_group = parser.add_argument_group('Exclusions')
    excluded_slicing_arguments(exc_group,project_drs,action_type='append')
    return

def validate(subparsers,epilog,project_drs):
    #Find Optimset Months
    epilog_validate=textwrap.dedent(epilog)
    parser=subparsers.add_parser('validate',
                                   description=textwrap.dedent('Take as an input the results from \'ask\'.\n\
                                         Returns pointers to models that have ALL the\n\
                                         requested experiments and variables for ALL requested years.\n\
                                         \n\
                                         Can be SLOW.\n\
                                         \n\
                                         Note that if this function fails it is likely that approriate\n\
                                         certificates have not been installed on this machine.'),
                                   epilog=epilog_validate,
                                 )
    input_arguments(parser)
    output_arguments(parser)
    proc_group = parser.add_argument_group('These arguments set threading options')
    proc_group.add_argument('--num_procs',
                                 default=1, type=int,#choices=xrange(1,6),
                                 help=textwrap.dedent('Use num_procs processes to perform the computation. This function might not work with your installation.'))
                                 #\n\
                                 #      Has been found to be rather unstable with more than 5 processses.'))

    data_node_group = parser.add_argument_group('Restrict search to specific data nodes')
    data_node_group.add_argument('--data_node',type=str,action='append',help='Consider only the specified data nodes')
    data_node_group.add_argument('--Xdata_node',type=str,action='append',help='Do not consider the specified data nodes')

    #inc_group = parser.add_argument_group('Inclusions')
    #slicing_arguments(inc_group,project_drs)
    #exc_group = parser.add_argument_group('Exclusions')
    #excluded_slicing_arguments(exc_group,project_drs)
    inc_group = parser.add_argument_group('Inclusions')
    slicing_arguments(inc_group,project_drs,action_type='append')
    exc_group = parser.add_argument_group('Exclusions')
    excluded_slicing_arguments(exc_group,project_drs,action_type='append')
    return

def download(subparsers,epilog,project_drs):
    epilog_validate=textwrap.dedent(epilog)
    parser=subparsers.add_parser('download',
                                           description=textwrap.dedent('Take as an input the results from \'validate\' and returns a\n\
                                                 netcdf file with the data retrieved.'),
                                           epilog=epilog_validate,
                                         )
    input_arguments(parser)
    output_arguments(parser)
    serial_group = parser.add_argument_group('Specify asynchronous behavior')
    serial_group.add_argument('--serial',default=False,action='store_true',help='Downloads the files serially.')

    source_group = parser.add_argument_group('Specify sources')
    source_group.add_argument('--source_dir',default=None,help='local cache of data retrieved using \'download_raw\'')

    inc_group = parser.add_argument_group('Inclusions')
    inc_group.add_argument('--year',
                                 default=None, type=int, nargs='*',
                                 help='Retrieve only this year.')
    inc_group.add_argument('--month',
                                 default=None, type=int, nargs='*',
                                 help='Retrieve only this month (1 to 12).')
    inc_group.add_argument('--day',
                                 default=None, type=int, nargs='*',
                                 help='Retrieve only this calendar day.')
    #slicing_arguments(inc_group,project_drs)
    #exc_group = parser.add_argument_group('Exclusions')
    #excluded_slicing_arguments(exc_group,project_drs)
    slicing_arguments(inc_group,project_drs,action_type='append')
    exc_group = parser.add_argument_group('Exclusions')
    excluded_slicing_arguments(exc_group,project_drs,action_type='append')

    data_node_group = parser.add_argument_group('Limit download from specific data nodes')
    data_node_group.add_argument('--data_node',type=str,action='append',help='Retrieve only from the specified data nodes')
    data_node_group.add_argument('--Xdata_node',type=str,action='append',help='Do not retrieve from the specified data nodes')

    return

def download_raw(subparsers,epilog,project_drs):
    epilog_download_raw=textwrap.dedent(epilog)
    parser=subparsers.add_parser('download_raw',
                                           description=textwrap.dedent('Take as an input the results from \'validate\' and downloads the data.'),
                                           epilog=epilog_download_raw,
                                         )
    input_arguments(parser)
    parser.add_argument('out_destination',
                             help='Destination directory for retrieval.')
    #proc_group = parser.add_argument_group('These arguments set threading options')
    #proc_group.add_argument('--num_procs',
    #                             default=1, type=int,
    #                             help='Use num_procs processes to perform the computation.')
    serial_group = parser.add_argument_group('Specify asynchronous behavior')
    serial_group.add_argument('--serial',default=False,action='store_true',help='Downloads the files serially.')

    source_group = parser.add_argument_group('Specify sources')
    source_group.add_argument('--source_dir',default=None,help='local cache of data retrieved using \'download_raw\'')

    data_node_group = parser.add_argument_group('Limit download from specific data nodes')
    data_node_group.add_argument('--data_node',type=str,action='append',help='Retrieve only from the specified data nodes')
    data_node_group.add_argument('--Xdata_node',type=str,action='append',help='Do not retrieve from the specified data nodes')

    inc_group = parser.add_argument_group('Inclusions')
    inc_group.add_argument('--year',
                                 default=None, type=int, nargs='*',
                                 help='Retrieve only this year.')
    inc_group.add_argument('--month',
                                 default=None, type=int, nargs='*',
                                 help='Retrieve only this month (1 to 12).')
    inc_group.add_argument('--day',
                                 default=None, type=int, nargs='*',
                                 help='Retrieve only this calendar day.')
    #slicing_arguments(inc_group,project_drs)
    #exc_group = parser.add_argument_group('Exclusions')
    #excluded_slicing_arguments(exc_group,project_drs)
    slicing_arguments(inc_group,project_drs,action_type='append')
    exc_group = parser.add_argument_group('Exclusions')
    excluded_slicing_arguments(exc_group,project_drs,action_type='append')
    return

def convert(subparsers,epilog,project_drs):
    epilog_convert=textwrap.dedent(epilog)
    parser=subparsers.add_parser('convert',
                                           description=textwrap.dedent('Take as an input the results from \'download\' and converts the data.'),
                                           epilog=epilog_convert,
                                         )
    input_arguments(parser)
    parser.add_argument('out_destination',
                             help='Destination directory for retrieval.')
    proc_group = parser.add_argument_group('These arguments set threading options')
    proc_group.add_argument('--num_procs',
                                 default=1, type=int,
                                 help='Use num_procs processes to perform the computation.')

    inc_group = parser.add_argument_group('Inclusions')
    slicing_arguments(inc_group,project_drs)
    exc_group = parser.add_argument_group('Exclusions')
    excluded_slicing_arguments(exc_group,project_drs)
    return


def apply(subparsers,epilog,project_drs):
    epilog_apply=textwrap.dedent(epilog)
    parser=subparsers.add_parser('apply',
                                           description=textwrap.dedent('Take as an input retrieved data and apply bash script'),
                                           epilog=epilog_apply
                                         )
    parser.add_argument('script',default='',help="Command-line script")
    parser.add_argument('in_diagnostic_netcdf_file',
                                 help='NETCDF retrieved files (input).')
    parser.add_argument('in_extra_netcdf_files',nargs='*',
                                 help='NETCDF extra retrieved files (input).')
    parser.add_argument('out_netcdf_file',
                                 help='NETCDF file (output)')

    select_group = parser.add_argument_group('These arguments specify the structure of the output')
    select_group.add_argument('--add_fixed',default=False, action='store_true',help='include fixed variables')
    select_group.add_argument('-f','--field',action='append', type=str, choices=project_drs.official_drs_no_version,
                                       help='Keep these fields in the applied file.' )
    proc_group = parser.add_argument_group('These arguments set threading options')
    proc_group.add_argument('--num_procs',
                                 default=1, type=int,
                                 help='Use num_procs processes to perform the computation.')

    inc_group = parser.add_argument_group('Inclusions')
    #slicing_arguments(inc_group,project_drs)
    #exc_group = parser.add_argument_group('Exclusions')
    #excluded_slicing_arguments(exc_group,project_drs)
    slicing_arguments(inc_group,project_drs,action_type='append')
    exc_group = parser.add_argument_group('Exclusions')
    excluded_slicing_arguments(exc_group,project_drs,action_type='append')
    return
    
